# cython: language_level=3, language=c++, embedsignature=True
cimport numpy as np

from cyarray.carray cimport BaseArray, UIntArray, IntArray, LongArray

# ParticleTag
# Declares various tags for particles, and functions to check them.

# Note that these tags are the ones set in the 'tag' property of the
# particles, in a particle array. To define additional discrete properties,
# one can add another integer property to the particles in the particle array
# while creating them.

# These tags could be considered as 'system tags' used internally to
# distinguish among different kinds of particles. If more tags are needed for
# a particular application, add them as mentioned above.

# The is_* functions defined below are to be used in Python for tests
# etc. Cython modules can directly use the enum name.

cdef enum ParticleTag:
    Local = 0
    Remote
    Ghost

cpdef bint is_local(int tag)
cpdef bint is_remote(int tag)
cpdef bint is_ghost(int tag)

cpdef int get_local_tag()
cpdef int get_remote_tag()
cpdef int get_ghost_tag()

cdef class ParticleArray:
    """
    Maintains various properties for particles.
    """
    cdef public str backend
    # dictionary to hold the properties held per particle
    cdef public dict properties
    cdef public list property_arrays

    cdef public dict stride

    # list of output property arrays
    cdef public list output_property_arrays

    # dictionary to hold the constants for all the particles
    cdef public dict constants

    # default value associated with each property
    cdef public dict default_values

    # name associated with this particle array
    cdef public str name

    # the number of real particles.
    cdef public long num_real_particles

    # a list of props to be used for load balancing
    cdef list lb_props

    ########################################
    # OpenCL/accelerator related attributes.

    # Object that manages the device properties.
    cdef public object gpu

    # time for the particle array
    cdef public double time

    cdef object _create_c_array_from_npy_array(self, np.ndarray arr)
    cdef _check_property(self, str)

    cdef np.ndarray _get_real_particle_prop(self, str prop)

    # set/get the time
    cpdef set_time(self, double time)
    cpdef double get_time(self)

    cpdef set_name(self, str name)

    cpdef get_lb_props(self)

    cpdef set_num_real_particles(self, long value)

    cpdef BaseArray get_carray(self, str prop)

    cpdef int get_number_of_particles(self, bint real=*)
    cpdef remove_particles(self, indices, align=*)
    cpdef remove_tagged_particles(self, int tag, bint align=*)

    # function to add any property
    cpdef add_constant(self, str name, data)
    cpdef add_property(self, str name, str type=*, default=*, data=*, stride=*)
    cpdef remove_property(self, str prop_name)

    # increase the number of particles by num_particles
    cpdef extend(self, int num_particles)

    cpdef has_array(self, str arr_name)

    # aligns all the real particles in contiguous positions starting from 0
    cpdef int align_particles(self) except -1

    # add particles from the parray to self.
    cpdef int append_parray(self, ParticleArray parray, bint align=*,
            bint update_constants=*) except -1

    cpdef ParticleArray empty_clone(self, props=*)

    cpdef ensure_properties(self, ParticleArray src, list props=*)

    # create a new particle array with the given particles indices and the
    # properties.
    cpdef ParticleArray extract_particles(self, indices, ParticleArray dest_array=*,
                            bint align=*, list props=*)

    # set the tag value for the particles
    cpdef set_tag(self, long tag_value, LongArray indices)

    cpdef copy_properties(self, ParticleArray source, long start_index=*, long
                          end_index=*)

    # copy properties from one set of variables to another
    cpdef copy_over_properties(self, dict props)

    # set the pid for all local particles
    cpdef set_pid(self, int pid)

    # set the specified properties to zero
    cpdef set_to_zero(self, list props)

    # resize all arrays to a new size
    cpdef resize(self, long size)



"""The basic equations for the IISPH formulation of

    M. Ihmsen, J. Cornelis, B. Solenthaler, C. Horvath, M. Teschner, "Implicit
    Incompressible SPH," IEEE Transactions on Visualization and Computer
    Graphics, vol. 20, no. 3, pp. 426-435, March 2014.
    http://dx.doi.org/10.1109/TVCG.2013.105

"""

from numpy import sqrt, fabs
from compyle.api import declare
from pysph.base.particle_array import get_ghost_tag
from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
from pysph.base.reduce_array import serial_reduce_array, parallel_reduce_array
from pysph.sph.scheme import Scheme, add_bool_argument


GHOST_TAG = get_ghost_tag()


class IISPHStep(IntegratorStep):
    """A straightforward and simple integrator to be used for IISPH.
    """
    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w,
               d_uadv, d_vadv, d_wadv, d_au, d_av, d_aw,
               d_ax, d_ay, d_az, dt):
        d_u[d_idx] = d_uadv[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_vadv[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_wadv[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] += dt * d_u[d_idx]
        d_y[d_idx] += dt * d_v[d_idx]
        d_z[d_idx] += dt * d_w[d_idx]


class NumberDensity(Equation):
    def initialize(self, d_idx, d_V):
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_V, WIJ):
        d_V[d_idx] += WIJ


class SummationDensity(Equation):
    def initialize(self, d_idx, d_rho):
        d_rho[d_idx] = 0.0

    def loop(self, d_idx, d_rho, s_idx, s_m, WIJ):
        d_rho[d_idx] += s_m[s_idx]*WIJ


class SummationDensityBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(SummationDensityBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, s_idx, s_V, WIJ):
        d_rho[d_idx] += self.rho0/s_V[s_idx]*WIJ


class NormalizedSummationDensity(Equation):
    def initialize(self, d_idx, d_rho, d_rho_adv, d_rho0, d_V):
        d_rho0[d_idx] = d_rho[d_idx]
        d_rho[d_idx] = 0.0
        d_rho_adv[d_idx] = 0.0
        d_V[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_rho_adv, d_V, s_idx, s_m, s_rho0, WIJ):
        tmp = s_m[s_idx]*WIJ
        d_rho[d_idx] += tmp
        d_rho_adv[d_idx] += tmp/s_rho0[s_idx]
        d_V[d_idx] += WIJ

    def post_loop(self, d_idx, d_rho, d_rho_adv):
        d_rho[d_idx] /= d_rho_adv[d_idx]


class AdvectionAcceleration(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(AdvectionAcceleration, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_uadv, d_vadv, d_wadv):
        d_au[d_idx] = self.gx
        d_av[d_idx] = self.gy
        d_aw[d_idx] = self.gz
        d_uadv[d_idx] = 0.0
        d_vadv[d_idx] = 0.0
        d_wadv[d_idx] = 0.0

    def post_loop(self, d_idx, d_au, d_av, d_aw, d_uadv, d_vadv, d_wadv,
                  d_u, d_v, d_w, dt=0.0):
        d_uadv[d_idx] = d_u[d_idx] + dt*d_au[d_idx]
        d_vadv[d_idx] = d_v[d_idx] + dt*d_av[d_idx]
        d_wadv[d_idx] = d_w[d_idx] + dt*d_aw[d_idx]


class ViscosityAcceleration(Equation):
    def __init__(self, dest, sources, nu):
        self.nu = nu
        super(ViscosityAcceleration, self).__init__(dest, sources)

    def loop(self, d_idx, d_au, d_av, d_aw, s_idx, s_m, EPS,
             VIJ, XIJ, RHOIJ1, R2IJ, DWIJ):
        dwijdotxij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]
        fac = 2.0*self.nu*s_m[s_idx]*RHOIJ1*dwijdotxij/(R2IJ + EPS)
        d_au[d_idx] += fac*VIJ[0]
        d_av[d_idx] += fac*VIJ[1]
        d_aw[d_idx] += fac*VIJ[2]


class ViscosityAccelerationBoundary(Equation):
    """The acceleration on the fluid due to a boundary.
    """
    def __init__(self, dest, sources, rho0, nu):
        self.nu = nu
        self.rho0 = rho0
        super(ViscosityAccelerationBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_au, d_av, d_aw, d_rho, s_idx, s_V, EPS,
             VIJ, XIJ, R2IJ, DWIJ):
        phi_b = self.rho0/(s_V[s_idx]*d_rho[d_idx])
        dwijdotxij = DWIJ[0]*XIJ[0] + DWIJ[1]*XIJ[1] + DWIJ[2]*XIJ[2]

        fac = 2.0*self.nu*phi_b*dwijdotxij/(R2IJ + EPS)
        d_au[d_idx] += fac*VIJ[0]
        d_av[d_idx] += fac*VIJ[1]
        d_aw[d_idx] += fac*VIJ[2]


class ComputeDII(Equation):
    def initialize(self, d_idx, d_dii0, d_dii1, d_dii2):
        d_dii0[d_idx] = 0.0
        d_dii1[d_idx] = 0.0
        d_dii2[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_dii0, d_dii1, d_dii2,
             s_idx, s_m, DWIJ):
        rho_1 = 1.0/d_rho[d_idx]
        fac = -s_m[s_idx]*rho_1*rho_1
        d_dii0[d_idx] += fac*DWIJ[0]
        d_dii1[d_idx] += fac*DWIJ[1]
        d_dii2[d_idx] += fac*DWIJ[2]


class ComputeDIIBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeDIIBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_dii0, d_dii1, d_dii2, d_rho,
             s_idx, s_m, s_V, DWIJ):
        rhoi1 = 1.0/d_rho[d_idx]
        fac = -rhoi1*rhoi1*self.rho0/s_V[s_idx]
        d_dii0[d_idx] += fac*DWIJ[0]
        d_dii1[d_idx] += fac*DWIJ[1]
        d_dii2[d_idx] += fac*DWIJ[2]


class ComputeRhoAdvection(Equation):
    def initialize(self, d_idx, d_rho_adv, d_rho, d_p0, d_p, d_piter, d_aii):
        d_rho_adv[d_idx] = d_rho[d_idx]
        d_p0[d_idx] = d_p[d_idx]
        d_piter[d_idx] = 0.5*d_p[d_idx]

    def loop(self, d_idx, d_rho, d_rho_adv, d_uadv, d_vadv, d_wadv, d_u,
             d_v, d_w, s_idx, s_m, s_uadv, s_vadv, s_wadv, DWIJ, dt=0.0):

        vijdotdwij = (d_uadv[d_idx] - s_uadv[s_idx])*DWIJ[0] + \
                     (d_vadv[d_idx] - s_vadv[s_idx])*DWIJ[1] + \
                     (d_wadv[d_idx] - s_wadv[s_idx])*DWIJ[2]

        d_rho_adv[d_idx] += dt*s_m[s_idx]*vijdotdwij


class ComputeRhoBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeRhoBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, d_rho_adv, d_uadv, d_vadv, d_wadv,
             s_idx, s_u, s_v, s_w, s_V, WIJ, DWIJ, dt=0.0):
        phi_b = self.rho0/s_V[s_idx]

        vijdotdwij = (d_uadv[d_idx] - s_u[s_idx])*DWIJ[0] + \
                     (d_vadv[d_idx] - s_v[s_idx])*DWIJ[1] + \
                     (d_wadv[d_idx] - s_w[s_idx])*DWIJ[2]
        d_rho_adv[d_idx] += dt*phi_b*vijdotdwij


class ComputeAII(Equation):
    def initialize(self, d_idx, d_aii):
        d_aii[d_idx] = 0.0

    def loop(self, d_idx, d_aii, d_dii0, d_dii1, d_dii2, d_m, d_rho,
             s_idx, s_m, s_rho, DWIJ):
        rho1 = 1.0/d_rho[d_idx]
        fac = d_m[d_idx]*rho1*rho1
        # The following is m_j (d_ii - d_ji) . DWIJ
        # DWIJ = -DWJI
        dijdotdwij = (d_dii0[d_idx] - fac*DWIJ[0])*DWIJ[0] + \
                     (d_dii1[d_idx] - fac*DWIJ[1])*DWIJ[1] + \
                     (d_dii2[d_idx] - fac*DWIJ[2])*DWIJ[2]
        d_aii[d_idx] += s_m[s_idx]*dijdotdwij


class ComputeAIIBoundary(Equation):
    """ This is important and not really discussed in the original IISPH paper.
    """
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(ComputeAIIBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_m, d_aii, d_dii0, d_dii1, d_dii2, d_rho,
             s_idx, s_m, s_V, DWIJ):
        phi_b = self.rho0/s_V[s_idx]
        rho1 = 1.0/d_rho[d_idx]
        fac = d_m[d_idx]*rho1*rho1
        dijdotdwij = ((d_dii0[d_idx] - fac*DWIJ[0])*DWIJ[0] +
                      (d_dii1[d_idx] - fac*DWIJ[1])*DWIJ[1] +
                      (d_dii2[d_idx] - fac*DWIJ[2])*DWIJ[2])
        d_aii[d_idx] += phi_b*dijdotdwij


class ComputeDIJPJ(Equation):
    def initialize(self, d_idx, d_dijpj0, d_dijpj1, d_dijpj2):
        d_dijpj0[d_idx] = 0.0
        d_dijpj1[d_idx] = 0.0
        d_dijpj2[d_idx] = 0.0

    def loop(self, d_idx, d_dijpj0, d_dijpj1, d_dijpj2,
             s_idx, s_m, s_rho, s_piter, DWIJ):
        rho1 = 1.0/s_rho[s_idx]
        fac = -s_m[s_idx]*rho1*rho1*s_piter[s_idx]
        d_dijpj0[d_idx] += fac*DWIJ[0]
        d_dijpj1[d_idx] += fac*DWIJ[1]
        d_dijpj2[d_idx] += fac*DWIJ[2]


class UpdateGhostProps(Equation):
    def __init__(self, dest, sources=None):
        super(UpdateGhostProps, self).__init__(dest, sources)
        # We do this to ensure that the ghost tag is indeed 2.
        # If not the initialize method will never work.
        assert GHOST_TAG == 2

    def initialize(self, d_idx, d_tag, d_orig_idx, d_dijpj0, d_dijpj1,
                   d_dijpj2, d_dii0, d_dii1, d_dii2, d_piter):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            d_dijpj0[d_idx] = d_dijpj0[idx]
            d_dijpj1[d_idx] = d_dijpj1[idx]
            d_dijpj2[d_idx] = d_dijpj2[idx]
            d_dii0[d_idx] = d_dii0[idx]
            d_dii1[d_idx] = d_dii1[idx]
            d_dii2[d_idx] = d_dii2[idx]
            d_piter[d_idx] = d_piter[idx]


class PressureSolve(Equation):
    def __init__(self, dest, sources, rho0, omega=0.5,
                 tolerance=1e-2, debug=False):
        self.rho0 = rho0
        self.omega = omega
        self.compression = 0.0
        self.debug = debug
        self.tolerance = tolerance
        super(PressureSolve, self).__init__(dest, sources)

    def initialize(self, d_idx, d_p, d_compression):
        d_p[d_idx] = 0.0
        d_compression[d_idx] = 0.0

    def loop(self, d_idx, d_p, d_piter, d_rho, d_m, d_dijpj0, d_dijpj1,
             d_dijpj2, s_idx, s_m, s_dii0, s_dii1, s_dii2,
             s_piter, s_dijpj0, s_dijpj1, s_dijpj2, DWIJ):

        # Note that a good way to check this is to see that when
        # d_idx == s_idx the contribution is zero, as is expected.
        rho1 = 1.0/d_rho[d_idx]
        fac = d_m[d_idx]*rho1*rho1*d_piter[d_idx]
        djkpk0 = s_dijpj0[s_idx] - fac*DWIJ[0]
        djkpk1 = s_dijpj1[s_idx] - fac*DWIJ[1]
        djkpk2 = s_dijpj2[s_idx] - fac*DWIJ[2]

        tmp0 = d_dijpj0[d_idx] - s_dii0[s_idx]*s_piter[s_idx] - djkpk0
        tmp1 = d_dijpj1[d_idx] - s_dii1[s_idx]*s_piter[s_idx] - djkpk1
        tmp2 = d_dijpj2[d_idx] - s_dii2[s_idx]*s_piter[s_idx] - djkpk2
        tmpdotdwij = tmp0*DWIJ[0] + tmp1*DWIJ[1] + tmp2*DWIJ[2]

        # This is corrected in the post_loop.
        d_p[d_idx] += s_m[s_idx]*tmpdotdwij

    def post_loop(self, d_idx, d_piter, d_p0, d_p, d_aii, d_rho_adv, d_rho,
                  d_compression, dt=0.0):
        dt2 = dt*dt
        # Recall that d_p now has \sum_{j\neq i} a_ij p_j
        tmp = self.rho0 - d_rho_adv[d_idx] - d_p[d_idx]*dt2
        dnr = d_aii[d_idx]*dt2

        if fabs(dnr) > 1e-9:
            # Clamp pressure to positive values.
            p = max((1.0 - self.omega)*d_piter[d_idx] +
                    self.omega/dnr*tmp, 0.0)
        else:
            p = 0.0

        if p != 0.0:
            d_compression[d_idx] = fabs(p*dnr - tmp) + self.rho0
        else:
            d_compression[d_idx] = self.rho0

        d_piter[d_idx] = p
        d_p[d_idx] = p

    def reduce(self, dst, t, dt):
        dst.tmp_comp[0] = serial_reduce_array(dst.compression > 0.0, 'sum')
        dst.tmp_comp[1] = serial_reduce_array(dst.compression, 'sum')
        dst.tmp_comp[:] = parallel_reduce_array(dst.tmp_comp, 'sum')
        if dst.tmp_comp[0] > 0:
            avg_rho = dst.tmp_comp[1]/dst.tmp_comp[0]
        else:
            avg_rho = self.rho0
        self.compression = fabs(avg_rho - self.rho0)/self.rho0

    def converged(self):
        debug = self.debug
        compression = self.compression

        if compression > self.tolerance:
            if debug:
                print("Not converged:", compression)
            return -1.0
        else:
            if debug:
                print("Converged:", compression)
            return 1.0


class PressureSolveBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(PressureSolveBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_p, d_rho, d_dijpj0, d_dijpj1, d_dijpj2,
             s_idx, s_V, DWIJ):
        phi_b = self.rho0/s_V[s_idx]
        dijdotwij = (d_dijpj0[d_idx]*DWIJ[0] +
                     d_dijpj1[d_idx]*DWIJ[1] +
                     d_dijpj2[d_idx]*DWIJ[2])
        d_p[d_idx] += phi_b*dijdotwij


class UpdateGhostPressure(Equation):
    def initialize(self, d_idx, d_tag, d_orig_idx, d_p, d_piter):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            d_piter[d_idx] = d_piter[idx]
            d_p[d_idx] = d_p[idx]


class PressureForce(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, d_rho, d_p, d_au, d_av, d_aw,
             s_idx, s_m, s_rho, s_p, DWIJ):
        rhoi1 = 1.0/d_rho[d_idx]
        rhoj1 = 1.0/s_rho[s_idx]
        fac = -s_m[s_idx]*(d_p[d_idx]*rhoi1*rhoi1 + s_p[s_idx]*rhoj1*rhoj1)
        d_au[d_idx] += fac*DWIJ[0]
        d_av[d_idx] += fac*DWIJ[1]
        d_aw[d_idx] += fac*DWIJ[2]

    def post_loop(self, d_idx, d_au, d_av, d_aw,
                  d_uadv, d_vadv, d_wadv, d_dt_cfl, d_dt_force):
        fac = d_au[d_idx]*d_au[d_idx] + d_av[d_idx]*d_av[d_idx] +\
                   d_aw[d_idx]*d_aw[d_idx]
        vmag = sqrt(d_uadv[d_idx]*d_uadv[d_idx] + d_vadv[d_idx]*d_vadv[d_idx] +
                    d_wadv[d_idx]*d_wadv[d_idx])
        d_dt_cfl[d_idx] = 2.0*vmag
        d_dt_force[d_idx] = 2.0*fac


class PressureForceBoundary(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(PressureForceBoundary, self).__init__(dest, sources)

    def loop(self, d_idx, d_rho, d_au, d_av, d_aw,  d_p, s_idx, s_V, DWIJ):
        rho1 = 1.0/d_rho[d_idx]
        fac = -d_p[d_idx]*rho1*rho1*self.rho0/s_V[s_idx]
        d_au[d_idx] += fac*DWIJ[0]
        d_av[d_idx] += fac*DWIJ[1]
        d_aw[d_idx] += fac*DWIJ[2]


class IISPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, rho0, nu=0.0,
                 gx=0.0, gy=0.0, gz=0.0, omega=0.5, tolerance=1e-2,
                 debug=False, has_ghosts=False):
        '''The IISPH scheme

        Parameters
        ----------

        fluids : list(str)
            List of names of fluid particle arrays.
        solids : list(str)
            List of names of solid particle arrays.
        dim: int
            Dimensionality of the problem.
        rho0 : float
            Density of fluid.
        nu : float
            Kinematic viscosity.
        gx, gy, gz : float
            Componenents of body acceleration (gravity, external forcing etc.)
        omega : float
            Relaxation parameter for relaxed-Jacobi iterations.
        tolerance: float
            Tolerance for the convergence of pressure iterations as a fraction.
        debug: bool
            Produce some debugging output on iterations.
        has_ghosts: bool
            The problem has ghost particles so add equations for those.
        '''
        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.rho0 = rho0
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.omega = omega
        self.tolerance = tolerance
        self.debug = debug
        self.has_ghosts = has_ghosts

    def add_user_options(self, group):
        group.add_argument(
            '--omega', action="store", type=float, dest="omega",
            default=None, help="Relaxation parameter for Jacobi iterations."
        )
        group.add_argument(
            '--tolerance', action='store', type=float, dest='tolerance',
            default=None,
            help='Tolerance for convergence of iterations as a fraction'
        )
        add_bool_argument(
            group, 'iisph-debug', dest='debug', default=None,
            help="Produce some debugging output on convergence of iterations."
        )

    def consume_user_options(self, options):
        vars = ['omega', 'tolerance', 'debug']
        data = dict((var, self._smart_getattr(options, var))
                    for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        """Configure the solver to be generated.

        This is to be called before `get_solver` is called.

        Parameters
        ----------

        dim : int
            Number of dimensions.
        kernel : Kernel instance.
            Kernel to use, if none is passed a default one is used.
        integrator_cls : pysph.sph.integrator.Integrator
            Integrator class to use, use sensible default if none is
            passed.
        extra_steppers : dict
            Additional integration stepper instances as a dict.
        **kw : extra arguments
            Any additional keyword args are passed to the solver instance.
        """
        from pysph.base.kernels import CubicSpline
        from pysph.sph.integrator import EulerIntegrator
        from pysph.solver.solver import Solver
        if kernel is None:
            kernel = CubicSpline(dim=self.dim)

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = IISPHStep()

        cls = integrator_cls if integrator_cls is not None else EulerIntegrator
        integrator = cls(**steppers)

        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def get_equations(self):
        from pysph.sph.equation import Group
        has_ghosts = self.has_ghosts
        equations = []
        if self.solids:
            g1 = Group(
                equations=[NumberDensity(dest=x, sources=[x])
                           for x in self.solids]
            )
            equations.append(g1)

        g2 = Group(
            equations=[SummationDensity(dest=x, sources=self.fluids)
                       for x in self.fluids],
            real=False
        )
        equations.append(g2)

        if self.solids:
            g3 = Group(
                equations=[
                    SummationDensityBoundary(
                        dest=x, sources=self.solids, rho0=self.rho0
                    )
                    for x in self.fluids],
                real=False
            )
            equations.append(g3)

        eq = []
        for fluid in self.fluids:
            eq.extend([
                AdvectionAcceleration(
                    dest=fluid, sources=None,
                    gx=self.gx, gy=self.gy, gz=self.gz
                ),
                ComputeDII(dest=fluid, sources=self.fluids)
            ])
            if self.nu > 0.0:
                eq.append(
                    ViscosityAcceleration(
                        dest=fluid, sources=self.fluids, nu=self.nu
                    )
                )

            if self.solids:
                if self.nu > 0.0:
                    eq.append(
                        ViscosityAccelerationBoundary(
                            dest=fluid, sources=self.solids, nu=self.nu,
                            rho0=self.rho0,
                        )
                    )
                eq.append(
                    ComputeDIIBoundary(dest=fluid, sources=self.solids,
                                       rho0=self.rho0)
                )

        g4 = Group(equations=eq, real=False)
        equations.append(g4)

        eq = []
        for fluid in self.fluids:
            eq.extend([
                ComputeRhoAdvection(dest=fluid, sources=self.fluids),
                ComputeAII(dest=fluid, sources=self.fluids),
            ])
            if self.solids:
                eq.extend([
                    ComputeRhoBoundary(dest=fluid, sources=self.solids,
                                       rho0=self.rho0),
                    ComputeAIIBoundary(dest=fluid, sources=self.solids,
                                       rho0=self.rho0),
                ])
        g5 = Group(equations=eq)
        equations.append(g5)

        sg1 = Group(
            equations=[
                ComputeDIJPJ(dest=x, sources=self.fluids)
                for x in self.fluids
            ]
        )
        eq = []
        for fluid in self.fluids:
            eq.append(
                PressureSolve(dest=fluid, sources=self.fluids,
                              rho0=self.rho0, tolerance=self.tolerance,
                              debug=self.debug)
            )
            if self.solids:
                eq.append(
                    PressureSolveBoundary(
                        dest=fluid, sources=self.solids, rho0=self.rho0,
                    )
                )
        sg2 = Group(equations=eq)

        if has_ghosts:
            ghost1 = Group(
                equations=[UpdateGhostProps(dest=x, sources=None)
                           for x in self.fluids],
                real=False
            )
            ghost2 = Group(
                equations=[UpdateGhostPressure(dest=x, sources=None)
                           for x in self.fluids],
                real=False
            )
            solver_eqs = [sg1, ghost1, sg2, ghost2]
        else:
            solver_eqs = [sg1, sg2]
        g6 = Group(
            equations=solver_eqs,
            iterate=True, max_iterations=30, min_iterations=2
        )
        equations.append(g6)

        eq = []
        for fluid in self.fluids:
            eq.append(
                PressureForce(dest=fluid, sources=self.fluids)
            )
            if self.solids:
                eq.append(
                    PressureForceBoundary(
                        dest=fluid, sources=self.solids, rho0=self.rho0
                    )
                )
        g7 = Group(equations=eq)
        equations.append(g7)

        return equations

    def setup_properties(self, particles, clean=True):
        """Setup the particle arrays so they have the right set of properties
        for this scheme.

        Parameters
        ----------

        particles : list
            List of particle arrays.

        clean : bool
            If True, removes any unnecessary properties.
        """
        from pysph.base.utils import get_particle_array_iisph
        dummy = get_particle_array_iisph()
        props = set(dummy.properties.keys())
        for pa in particles:
            self._ensure_properties(pa, props, clean)
            for c, v in dummy.constants.items():
                if c not in pa.constants:
                    pa.add_constant(c, v)
            pa.set_output_arrays(dummy.output_property_arrays)

        if self.has_ghosts:
            particle_arrays = dict([(p.name, p) for p in particles])
            for fluid in self.fluids:
                pa = particle_arrays[fluid]
                pa.add_property('orig_idx', type='int')



from pysph.sph.wc.linalg import (
    augmented_matrix, gj_solve, mat_mult, mat_vec_mult
)
import numpy as np
import unittest


def gj_solve_helper(a, b, n):
    m = np.zeros((n, n+1)).ravel().tolist()
    augmented_matrix(a, b, n, 1, n, m)
    result = [0.0]*n
    is_singular = gj_solve(m, n, 1, result)
    return is_singular, result


class TestLinalg(unittest.TestCase):

    def _to_array(self, x, shape=None):
        x = np.asarray(x)
        if shape:
            x.shape = shape
        return x

    def test_augmented_matrix(self):
        # Given
        a = np.random.random((3, 3))
        b = np.random.random((3, 2))
        res = np.zeros((3, 5)).ravel().tolist()
        expect = np.zeros((3, 5))
        expect[:, :3] = a
        expect[:, 3:] = b
        # When
        augmented_matrix(a.ravel(), b.ravel(), 3, 2, 3, res)
        res = self._to_array(res, (3, 5))
        # Then
        np.testing.assert_array_almost_equal(res, expect)

    def test_augmented_matrix_with_lower_dimension(self):
        # Given
        a = np.random.random((3, 3))
        b = np.random.random((3, 2))
        res = np.zeros((3, 5)).ravel().tolist()
        expect = np.zeros((2, 4))
        expect[:, :2] = a[:2, :2]
        expect[:, 2:] = b[:2, :]
        expect.resize((3, 5), refcheck=False)
        # When
        augmented_matrix(a.ravel(), b.ravel(), 2, 2, 3, res)
        res = self._to_array(res, (3, 5))
        # Then
        np.testing.assert_array_almost_equal(res, expect)

    def test_augmented_matrix_with_gjsolve_with_lower_dimension(self):
        # Given
        nmax = 3
        mat = np.array([[7., 4., 2.], [8., 9., 4.], [1., 4., 10.]])
        b = np.array([5., 4., 2.])
        expect = np.linalg.solve(mat[:2, :2], b[:2])
        augmat = np.zeros((3, 4)).ravel().tolist()
        res = np.zeros(2).ravel().tolist()
        # When
        augmented_matrix(mat.ravel(), b.ravel(), 2, 1, nmax, augmat)
        gj_solve(augmat, 2, 1, res)
        # Then
        np.testing.assert_array_almost_equal(res, expect)

    def test_general_matrix(self):
        # Test Gauss Jordan solve.
        """
        This is a general matrix which needs partial pivoting to be
        solved.
        References
        ----------
        http://web.mit.edu/10.001/Web/Course_Notes/GaussElimPivoting.html
        """
        n = 4
        mat = [[0.02, 0.01, 0., 0.], [1., 2., 1., 0.], [0., 1., 2., 1.],
               [0., 0., 100., 200.]]
        b = [0.02, 1., 4., 800.]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_band_matrix(self):
        n = 3
        mat = [[1., -2., 0.], [1., -1., 3.], [2., 5., 0.]]
        b = [-3., 1., 0.5]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_dense_matrix(self):
        n = 3
        mat = [[0.96, 4.6, -3.7], [2.7, 4.3, -0.67], [0.9, 0., -5.]]
        b = [2.4, 3.6, -5.8]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_tridiagonal_matrix(self):
        n = 4
        mat = [[-2., 1., 0., 0.], [1., -2., 1., 0.], [0., 1., -2., 0.],
               [0., 0., 1., -2.]]
        b = [-1., 0., 0., -5.]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_symmetric_matrix(self):
        n = 3
        mat = [[0.96, 4.6, -3.7], [4.6, 4.3, -0.67], [-3.7, -0.67, -5.]]
        b = [2.4, 3.6, -5.8]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_symmetric_positivedefinite_Matrix(self):
        n = 4
        mat = [[1., 1., 4., -1.], [1., 5., 0., -1.], [4., 0., 21., -4.],
               [-1., -1., -4., 10.]]
        b = [2.4, 3.6, -5.8, 0.5]
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        mat = np.array(mat)
        new_b = np.dot(mat, np.transpose(np.array(result)))
        new_b = np.ravel(np.array(new_b))
        assert np.allclose(new_b, np.array(b))
        self.assertAlmostEqual(sing, 0.0)

    def test_inverse(self):
        # Given
        n = 3
        mat = [[1.0, 2.0, 2.5], [2.5, 1.0, 0.0], [0.0, 0.0, 1.0]]
        b = np.identity(3).ravel().tolist()
        A = np.zeros((3, 6)).ravel().tolist()
        augmented_matrix(np.ravel(mat), b, 3, 3, 3, A)
        result = np.zeros((3, 3)).ravel().tolist()

        # When
        sing = gj_solve(A, n, n, result)

        # Then
        mat = np.asarray(mat)
        res = np.asarray(result)
        res.shape = 3, 3
        np.testing.assert_allclose(res, np.linalg.inv(mat))
        self.assertAlmostEqual(sing, 0.0)

    def test_matmult(self):
        # Given
        n = 3
        a = np.random.random((3, 3))
        b = np.random.random((3, 3))
        result = [0.0]*9

        # When
        mat_mult(a.ravel(), b.ravel(), n, result)

        # Then.
        expect = np.dot(a, b)
        result = np.asarray(result)
        result.shape = 3, 3
        np.testing.assert_allclose(result, expect)

    def test_mat_vec_mult(self):
        # Given
        n = 3
        a = np.random.random((3, 3))
        b = np.random.random((3,))
        result = [0.0]*3

        # When
        mat_vec_mult(a.ravel(), b, n, result)

        # Then.
        expect = np.dot(a, b)
        result = np.asarray(result)
        np.testing.assert_allclose(result, expect)

    def test_singular_matrix(self):
        # Given
        n = 3
        mat = [[1., 1., 0.], [1., 1., 0.], [1., 1., 1.]]
        b = [1.0, 1.0, 1.0]
        #
        sing, result = gj_solve_helper(np.ravel(mat), b, n)
        self.assertAlmostEqual(sing, 1.0)


if __name__ == '__main__':
    unittest.main()



# Author: Prabhu Ramachandran
# Copyright (c) 2014 Prabhu Ramachandran
# License: BSD Style.

# Standard library imports
try:
    # This is for Python-2.6.x
    import unittest2 as unittest
except ImportError:
    import unittest

# Library imports.
import numpy as np

# Local imports
from pysph.tools.interpolator import get_nx_ny_nz, Interpolator
from pysph.base.nnps_base import DomainManager
from pysph.base.utils import get_particle_array


class TestGetNxNyNz(unittest.TestCase):
    def test_should_work_for_1d_data(self):
        # When
        num_points = 100
        bounds = (0.0, 1.5, 0.0, 0.0, 0.0, 0.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [100, 1, 1])

        # When
        num_points = 100
        bounds = (0.0, 0.0, 0.0, 1.123, 0.0, 0.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [1, 100, 1])

        # When
        num_points = 100
        bounds = (0.0, 0.0, 0.0, 0.0, 0.0, 3.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [1, 1, 100])

    def test_should_work_for_2d_data(self):
        # When
        num_points = 100
        bounds = (0.0, 1.0, 0.5, 1.5, 0.0, 0.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [10, 10, 1])

        # When
        num_points = 100
        bounds = (0.0, 0.0, 0, 1, 0.5, 1.5)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [1, 10, 10])

    def test_should_work_for_3d_data(self):
        # When
        num_points = 1000
        bounds = (0.0, 1.0, 0.5, 1.5, -1.0, 0.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [10, 10, 10])

        # When
        num_points = 1000
        bounds = (0.0, 1.0, 0.5, 1.5, -1.0, 2.0)
        dims = get_nx_ny_nz(num_points, bounds)

        # Then
        self.assertListEqual(list(dims), [7, 7, 21])


class TestInterpolator(unittest.TestCase):
    def _make_2d_grid(self, name='fluid'):
        n = 101
        dx = 2.0/(n-1)
        x, y = np.mgrid[-1.+dx/2:1.:dx, -1.+dx/2:1.:dx]
        z = np.zeros_like(x)
        x, y, z = x.ravel(), y.ravel(), z.ravel()
        m = np.ones_like(x)
        p = np.sin(x * np.pi)
        h = np.ones_like(x)*2.*dx
        u = np.cos(x * np.pi)
        rho = np.ones_like(x)*m/(dx*dx)
        pa = get_particle_array(name=name, x=x, y=y, z=z, h=h, m=m, p=p, u=u,
                                rho=rho)
        return pa

    @property
    def _domain(self):
        x0 = 1.0
        domain = DomainManager(
            xmin=-x0, xmax=x0, ymin=-x0, ymax=x0,
            periodic_in_x=True, periodic_in_y=True
        )
        return domain

    def test_should_work_on_2d_data(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)
        p = ip.interpolate('p')
        u = ip.interpolate('u')

        # Then.
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)
        expect = np.cos(ip.x * np.pi)
        np.testing.assert_allclose(u, expect, rtol=5e-3)

    def test_should_work_with_multiple_arrays(self):
        # Given
        pa1 = self._make_2d_grid()
        pa2 = self._make_2d_grid('solid')
        pa2.p[:] = 4.0
        pa2.u[:] = 0.2

        # When.
        ip = Interpolator([pa1, pa2], num_points=1000,
                          domain_manager=self._domain)
        p = ip.interpolate('p')
        u = ip.interpolate('u')

        # Then.
        expect = (np.sin(ip.x * np.pi) + 4.0) * 0.5
        np.testing.assert_allclose(p, expect, rtol=5e-2)
        expect = (np.cos(ip.x * np.pi) + 0.2) * 0.5
        np.testing.assert_allclose(u, expect, rtol=5e-2)

    def test_should_work_with_ghost_particles(self):
        # Given
        pa = self._make_2d_grid()
        # Make half the particles ghosts.
        n = pa.get_number_of_particles()
        pa.tag[int(n//2):] = 1
        pa.align_particles()

        # When.
        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)
        p = ip.interpolate('p')
        u = ip.interpolate('u')

        # Then.
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)
        expect = np.cos(ip.x * np.pi)
        np.testing.assert_allclose(u, expect, rtol=5e-3)

    def test_should_work_with_changed_data(self):
        # Given
        pa = self._make_2d_grid()
        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)
        p = ip.interpolate('p')

        # When.
        pa.p *= 2.0
        ip.update()
        p = ip.interpolate('p')

        # Then.
        expect = np.sin(ip.x * np.pi) * 2.0
        np.testing.assert_allclose(p, expect, rtol=5e-2)

    def test_should_be_able_to_update_particle_arrays(self):
        # Given
        pa = self._make_2d_grid()
        pa_new = self._make_2d_grid()
        pa_new.p[:] = np.cos(pa_new.x * np.pi)

        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)
        p = ip.interpolate('p')

        # When.
        ip.update_particle_arrays([pa_new])
        p = ip.interpolate('p')

        # Then.
        expect = np.cos(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_should_correctly_update_domain(self):
        # Given
        pa = self._make_2d_grid()
        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)
        p = ip.interpolate('p')

        # When.
        ip.set_domain((0.1, 1.0, 0.1, 1.0, 0.0, 0.0), (11, 11, 1))
        p = ip.interpolate('p')

        # Then.
        expect = np.sin(ip.x * np.pi)
        print(p - expect)
        np.testing.assert_allclose(p, expect, atol=5e-3)

    def test_should_work_when_arrays_have_different_props(self):
        # Given
        pa1 = self._make_2d_grid()
        pa1.add_property('junk', default=2.0, type='int')
        pa2 = self._make_2d_grid('solid')

        # When.
        ip = Interpolator([pa1, pa2], num_points=1000,
                          domain_manager=self._domain)
        junk = ip.interpolate('junk')

        # Then.
        expect = np.ones_like(junk)*1.0
        self.assertTrue(np.allclose(junk, expect))

    def test_should_work_with_explicit_points_in_constructor(self):
        # Given
        pa = self._make_2d_grid()
        x, y = np.random.random((2, 5, 5))
        z = np.zeros_like(x)

        # When.
        ip = Interpolator([pa], x=x, y=y, z=z, domain_manager=self._domain)
        p = ip.interpolate('p')

        # Then.
        self.assertEqual(p.shape, x.shape)
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_should_work_with_explicit_points_without_z(self):
        # Given
        pa = self._make_2d_grid()
        x, y = np.random.random((2, 5, 5))

        # When.
        ip = Interpolator([pa], x=x, y=y, domain_manager=self._domain)
        p = ip.interpolate('p')

        # Then.
        self.assertEqual(p.shape, x.shape)
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_that_set_interpolation_points_works(self):
        # Given
        pa = self._make_2d_grid()
        ip = Interpolator([pa], num_points=1000, domain_manager=self._domain)

        # When.
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)
        p = ip.interpolate('p')

        # Then.
        self.assertEqual(p.shape, x.shape)
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_should_work_with_method_order1(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator(
            [pa], domain_manager=self._domain, method='order1'
        )
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)
        p = ip.interpolate('p')

        # Then.
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_should_work_with_method_sph(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator(
            [pa], domain_manager=self._domain, method='sph'
        )
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)
        p = ip.interpolate('p')

        # Then.
        expect = np.sin(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=5e-3)

    def test_gradient_calculation_2d(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator(
            [pa], domain_manager=self._domain, method='order1'
        )
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)
        p = ip.interpolate('p', 1)

        # Then.
        expect = np.pi * np.cos(ip.x * np.pi)
        np.testing.assert_allclose(p, expect, rtol=1e-2)

    def test_gradient_with_sph(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator(
            [pa], domain_manager=self._domain, method='sph'
        )
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)

        # Then.
        self.assertRaises(RuntimeError, ip.interpolate, 'p', 1)

    def test_gradient_with_shepard(self):
        # Given
        pa = self._make_2d_grid()

        # When.
        ip = Interpolator(
            [pa], domain_manager=self._domain, method='shepard'
        )
        x, y = np.random.random((2, 5, 5))
        ip.set_interpolation_points(x=x, y=y)

        # Then.
        self.assertRaises(RuntimeError, ip.interpolate, 'p', 1)


if __name__ == '__main__':
    unittest.main()



"""
References
-----------

    .. [Hopkins2013] Hopkins, Philip F. “A General Class of Lagrangian
        Smoothed Particle Hydrodynamics Methods and Implications for Fluid
        Mixing Problems.” Monthly Notices of the Royal Astronomical Society
        428, no. 4 (February 1, 2013): 2840–56.
        https://doi.org/10.1093/mnras/sts210.

    .. [Hopkins2015] Hopkins, Philip F. “A New Class of Accurate,
        Mesh-Free Hydrodynamic Simulation Methods.” Monthly Notices of the
        Royal Astronomical Society 450, no. 1 (June 11, 2015): 53–110.
        https://doi.org/10.1093/mnras/stv195.
"""
from compyle.types import declare
from pysph.base.particle_array import get_ghost_tag

from pysph.sph.equation import Equation
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.scheme import Scheme
from pysph.sph.wc.linalg import augmented_matrix, gj_solve, identity, mat_mult

GHOST_TAG = get_ghost_tag()


class TSPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, gamma, hfact, beta=2.0, fkern=1.0,
                 max_density_iterations=250, alphamax=1.0,
                 density_iteration_tolerance=1e-3, has_ghosts=False):
        """
        Density-energy formulation [Hopkins2013]_ including Balsara's
        artificial viscocity switch with modifications,
        as presented in Appendix F1 of [Hopkins2015]_ .

        Notes
        -----
        Is this exactly in accordance with what is proposed in [Hopkins2015]_ ?
            Not quite.

        What is different then?
            #. Adapting smoothing length using MPM [KP14]_ procedure from
               :class:`SummationDensity
               <pysph.sph.gas_dynamics.basic.SummationDensity>`. In this,
               calculation of grad-h terms are changed to that specified for
               this scheme.
            #. Using the PEC integrator step. No individual
               adaptive time-stepping.
            #. Using :class:`Gaussian Kernel <pysph.base.kernels.Gaussian>`
               by default instead of Cubic Spline with radius scale 1.

        Tip: Reduce the number of points if particle penetration is
        encountered. This has to be done while running
        ``gas_dynamics.wc_blastwave`` and ``gas_dynamics.robert``.

        Parameters
        ----------
        fluids: list
            List of names of fluid particle arrays.
        solids: list
            List of names of solid particle arrays (or boundaries), currently
            not supported
        dim: int
            Dimensionality of the problem.
        gamma: float
            :math:`\\gamma` for Equation of state.
        hfact: float
            :math:`h_{fact}` for smoothing length adaptivity, also referred to
            as kernel_factor in other gas dynamics schemes.
        beta : float, optional
            :math:`\\beta` for artificial viscosity, by default 2.0
        fkern : float, optional
            :math:`f_{kern}`, Factor to scale smoothing length for equivalence
            with classic kernel when using kernel with altered
            `radius_scale` is being used, by default 1.
        max_density_iterations : int, optional
            Maximum number of iterations to run for one density step,
            by default 250.
        density_iteration_tolerance : float, optional
            Maximum difference allowed in two successive density iterations,
            by default 1e-3
        has_ghosts : bool, optional
            if ghost particles (either mirror or periodic) is used, by default
            False
        alphamax : float, optional
            :math:`\\alpha_{av}` for artificial viscosity switch, by default
            1.0
        """

        self.fluids = fluids
        self.solids = solids
        self.dim = dim
        self.solver = None
        self.gamma = gamma
        self.beta = beta
        self.hfact = hfact
        self.density_iteration_tolerance = density_iteration_tolerance
        self.max_density_iterations = max_density_iterations
        self.has_ghosts = has_ghosts
        self.fkern = fkern
        self.alphamax = alphamax

    def add_user_options(self, group):
        group.add_argument("--alpha-max", action="store", type=float,
                           dest="alphamax", default=None,
                           help="alpha_max for the artificial viscosity "
                                "switch.")

        group.add_argument("--beta", action="store", type=float, dest="beta",
                           default=None,
                           help="beta for the artificial viscosity.")

        group.add_argument("--gamma", action="store", type=float, dest="gamma",
                           default=None, help="gamma for the state equation.")

    def consume_user_options(self, options):
        vars = ['gamma', 'alphamax', 'beta']
        data = dict((var, self._smart_getattr(options, var)) for var in vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):

        from pysph.base.kernels import Gaussian
        if kernel is None:
            kernel = Gaussian(dim=self.dim)

        if hasattr(kernel, 'fkern'):
            self.fkern = kernel.fkern
        else:
            self.fkern = 1.0

        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        from pysph.sph.integrator import PECIntegrator

        cls = integrator_cls if integrator_cls is not None else PECIntegrator
        step_cls = PECStep
        for name in self.fluids:
            if name not in steppers:
                steppers[name] = step_cls()

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(dim=self.dim, integrator=integrator,
                             kernel=kernel, **kw)

    def get_equations(self):
        from pysph.sph.equation import Group

        all_pa = self.fluids + self.solids
        equations = []

        # Find the optimal 'h'
        g1 = []
        for fluid in self.fluids:
            g1.append(SummationDensity(
                dest=fluid, sources=all_pa, hfact=self.hfact,
                density_iterations=True, dim=self.dim,
                htol=self.density_iteration_tolerance))
            equations.append(Group(
                equations=g1, update_nnps=True, iterate=True,
                max_iterations=self.max_density_iterations))

        g2 = []
        for fluid in self.fluids:
            g2.append(IdealGasEOS(dest=fluid, sources=None, gamma=self.gamma))
        equations.append(Group(equations=g2))

        g3 = []
        for fluid in self.fluids:
            g3.append(VelocityGradDivC1(dest=fluid, sources=all_pa,
                                        dim=self.dim))
            g3.append(BalsaraSwitch(dest=fluid, sources=None,
                                    alphaav=self.alphamax, fkern=self.fkern))
        equations.append(Group(equations=g3))

        g4 = []
        for solid in self.solids:
            g4.append(WallBoundary(solid, sources=self.fluids))
        equations.append(Group(equations=g4))

        if self.has_ghosts:
            gh = []
            for fluid in self.fluids:
                gh.append(UpdateGhostProps(dest=fluid, sources=None))
            equations.append(Group(equations=gh, real=False))

        g5 = []
        for fluid in self.fluids:
            g5.append(MomentumAndEnergy(dest=fluid, sources=all_pa,
                                        dim=self.dim, beta=self.beta,
                                        fkern=self.fkern))
        equations.append(Group(equations=g5))

        return equations

    def setup_properties(self, particles, clean=True):
        import numpy
        particle_arrays = dict([(p.name, p) for p in particles])

        props = ['rho', 'm', 'x', 'y', 'z', 'u', 'v', 'w', 'h', 'cs', 'p', 'e',
                 'au', 'av', 'aw', 'ae', 'pid', 'gid', 'tag', 'dwdh', 'h0',
                 'converged', 'ah', 'arho', 'dt_cfl', 'e0', 'rho0', 'u0', 'v0',
                 'w0', 'x0', 'y0', 'z0', 'alpha']
        more_props = ['drhosumdh', 'n', 'dndh', 'prevn', 'prevdndh',
                      'prevdrhosumdh', 'divv', 'an', 'n0']
        props.extend(more_props)
        output_props = 'rho p u v w x y z e n divv h alpha'.split(' ')
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.add_property('orig_idx', type='int')
            # Guess for number density.
            pa.add_property('n', data=pa.rho / pa.m)
            pa.add_property('gradv', stride=9)
            pa.add_property('invtt', stride=9)
            nfp = pa.get_number_of_particles()
            pa.orig_idx[:] = numpy.arange(nfp)
            pa.set_output_arrays(output_props)

        solid_props = set(props) | set('m0 wij htmp'.split(' '))
        for solid in self.solids:
            pa = particle_arrays[solid]
            self._ensure_properties(pa, solid_props, clean)
            pa.set_output_arrays(output_props)


class SummationDensity(Equation):
    def __init__(self, dest, sources, dim, density_iterations=False,
                 iterate_only_once=False, hfact=1.2, htol=1e-6):
        """
        :class:`SummationDensity
        <pysph.sph.gas_dynamics.basic.SummationDensity>` modified to use
        number density for calculation of grad-h terms.

        Ref. Appendix F1 [Hopkins2015]_
        """

        self.density_iterations = density_iterations
        self.iterate_only_once = iterate_only_once
        self.dim = dim
        self.hfact = hfact
        self.htol = htol
        self.equation_has_converged = 1

        super().__init__(dest, sources)

    def initialize(self, d_idx, d_rho, d_arho, d_drhosumdh, d_n, d_dndh,
                   d_prevn, d_prevdndh, d_prevdrhosumdh, d_an):

        d_rho[d_idx] = 0.0
        d_arho[d_idx] = 0.0

        d_prevn[d_idx] = d_n[d_idx]
        d_prevdrhosumdh[d_idx] = d_drhosumdh[d_idx]
        d_prevdndh[d_idx] = d_dndh[d_idx]

        d_drhosumdh[d_idx] = 0.0
        d_n[d_idx] = 0.0
        d_an[d_idx] = 0.0
        d_dndh[d_idx] = 0.0

        # set the converged attribute for the Equation to True. Within
        # the post-loop, if any particle hasn't converged, this is set
        # to False. The Group can therefore iterate till convergence.
        self.equation_has_converged = 1

    def loop(self, d_idx, s_idx, d_rho, d_arho, d_drhosumdh, s_m, VIJ, WI, DWI,
             GHI, d_n, d_dndh, d_h, d_prevn, d_prevdndh, d_prevdrhosumdh,
             d_an):

        mj = s_m[s_idx]
        vijdotdwij = VIJ[0] * DWI[0] + VIJ[1] * DWI[1] + VIJ[2] * DWI[2]

        # density
        d_rho[d_idx] += mj * WI

        # density accelerations
        hibynidim = d_h[d_idx] / (d_prevn[d_idx] * self.dim)
        inbrkti = 1 + d_prevdndh[d_idx] * hibynidim
        inprthsi = d_prevdrhosumdh[d_idx] * hibynidim
        fij = 1 - inprthsi / (s_m[s_idx] * inbrkti)
        vijdotdwij_fij = vijdotdwij * fij
        d_arho[d_idx] += mj * vijdotdwij_fij
        d_an[d_idx] += vijdotdwij_fij

        # gradient of kernel w.r.t h
        d_drhosumdh[d_idx] += mj * GHI
        d_n[d_idx] += WI
        d_dndh[d_idx] += GHI

    def post_loop(self, d_idx, d_h0, d_h, d_ah, d_converged, d_n, d_dndh,
                  d_an):
        # iteratively find smoothing length consistent with the
        if self.density_iterations:
            if not (d_converged[d_idx] == 1):
                hi = d_h[d_idx]
                hi0 = d_h0[d_idx]

                # estimated, without summations
                ni = (self.hfact / hi) ** self.dim
                dndhi = - self.dim * d_n[d_idx] / hi

                # the non-linear function and it's derivative
                func = d_n[d_idx] - ni
                dfdh = d_dndh[d_idx] - dndhi

                # Newton Raphson estimate for the new h
                hnew = hi - func / dfdh

                # Nanny control for h
                if hnew > 1.2 * hi:
                    hnew = 1.2 * hi
                elif hnew < 0.8 * hi:
                    hnew = 0.8 * hi

                # check for convergence
                diff = abs(hnew - hi) / hi0

                if not ((diff < self.htol) or self.iterate_only_once):
                    # this particle hasn't converged. This means the
                    # entire group must be repeated until this fellow
                    # has converged, or till the maximum iteration has
                    # been reached.
                    self.equation_has_converged = -1

                    # set particle properties for the next
                    # iteration. For the 'converged' array, a value of
                    # 0 indicates the particle hasn't converged
                    d_h[d_idx] = hnew
                    d_converged[d_idx] = 0
                else:
                    d_ah[d_idx] = d_an[d_idx] / dndhi
                    d_converged[d_idx] = 1

    def converged(self):
        return self.equation_has_converged


class IdealGasEOS(Equation):
    def __init__(self, dest, sources, gamma):
        """
        :class:`IdealGasEOS
        <pysph.sph.gas_dynamics.basic.IdealGasEOS>` modified to avoid repeated
        calculations using :meth:`loop() <pysph.sph.equation.Equation.loop()>`.
        Doing the same using :meth:`post_loop()
        <pysph.sph.equation.Equation.loop()>`.
        """
        self.gamma = gamma
        self.gamma1 = gamma - 1.0
        super(IdealGasEOS, self).__init__(dest, sources)

    def post_loop(self, d_idx, d_p, d_rho, d_e, d_cs):
        d_p[d_idx] = self.gamma1 * d_rho[d_idx] * d_e[d_idx]
        d_cs[d_idx] = sqrt(self.gamma * d_p[d_idx] / d_rho[d_idx])


class VelocityGradDivC1(Equation):
    def __init__(self, dest, sources, dim):
        """
        First Order consistent velocity gradient and divergence
        """
        self.dim = dim
        super().__init__(dest, sources)

    def _get_helpers_(self):
        return [augmented_matrix, gj_solve, identity, mat_mult]

    def initialize(self, d_gradv, d_idx, d_invtt, d_divv):
        start_indx, i, dim = declare('int', 3)
        start_indx = 9 * d_idx

        for i in range(9):
            d_gradv[start_indx + i] = 0.0
            d_invtt[start_indx + i] = 0.0

        d_divv[d_idx] = 0.0

    def loop(self, d_idx, d_invtt, s_m, s_idx, VIJ, DWI, XIJ, d_gradv):
        start_indx, row, col, drowcol, dim = declare('int', 5)
        dim = self.dim
        start_indx = d_idx * 9
        for row in range(dim):
            for col in range(dim):
                drowcol = start_indx + row * 3 + col
                d_invtt[drowcol] -= s_m[s_idx] * XIJ[row] * DWI[col]
                d_gradv[drowcol] -= s_m[s_idx] * VIJ[row] * DWI[col]

    def post_loop(self, d_idx, d_gradv, d_invtt, d_divv):
        tt, invtt, idmat, gradv = declare('matrix(9)', 4)
        augtt = declare('matrix(18)')

        start_indx, row, col, rowcol, drowcol, dim = declare('int', 6)

        dim = self.dim
        start_indx = 9 * d_idx
        identity(idmat, 3)
        identity(tt, 3)

        for row in range(3):
            for col in range(3):
                rowcol = row * 3 + col
                drowcol = start_indx + rowcol
                gradv[rowcol] = d_gradv[drowcol]

        for row in range(dim):
            for col in range(dim):
                rowcol = row * 3 + col
                drowcol = start_indx + rowcol
                tt[rowcol] = d_invtt[drowcol]

        augmented_matrix(tt, idmat, 3, 3, 3, augtt)
        gj_solve(augtt, 3, 3, invtt)
        gradvls = declare('matrix(9)')
        mat_mult(gradv, invtt, 3, gradvls)

        for row in range(dim):
            d_divv[d_idx] += gradvls[row * 3 + row]
            for col in range(dim):
                rowcol = row * 3 + col
                drowcol = start_indx + rowcol
                d_gradv[drowcol] = gradvls[rowcol]


class BalsaraSwitch(Equation):
    def __init__(self, dest, sources, alphaav, fkern):
        self.alphaav = alphaav
        self.fkern = fkern
        super().__init__(dest, sources)

    def post_loop(self, d_h, d_idx, d_cs, d_divv, d_gradv, d_alpha):
        curlv = declare('matrix(3)')

        curlv[0] = (d_gradv[9 * d_idx + 3 * 2 + 1] -
                    d_gradv[9 * d_idx + 3 * 1 + 2])
        curlv[1] = (d_gradv[9 * d_idx + 3 * 0 + 2] -
                    d_gradv[9 * d_idx + 3 * 2 + 0])
        curlv[2] = (d_gradv[9 * d_idx + 3 * 1 + 0] -
                    d_gradv[9 * d_idx + 3 * 0 + 1])

        abscurlv = sqrt(curlv[0] * curlv[0] +
                        curlv[1] * curlv[1] +
                        curlv[2] * curlv[2])

        absdivv = abs(d_divv[d_idx])

        fhi = d_h[d_idx] * self.fkern

        d_alpha[d_idx] = self.alphaav * absdivv / (
                absdivv + abscurlv + 0.0001 * d_cs[d_idx] / fhi)


class MomentumAndEnergy(Equation):
    def __init__(self, dest, sources, dim, fkern, beta=2.0):
        r"""
        TSPH Momentum and Energy Equations with artificial viscosity.

        Possible typo in that has been taken care of:

        Instead of Equation F3 [Hopkins2015]_ for evolution of total
        energy sans artificial viscosity and artificial conductivity,

            .. math::
                \frac{\mathrm{d} E_{i}}{\mathrm{~d} t}=\boldsymbol{v}_{i}
                \cdot \frac{\mathrm{d} \boldsymbol{P}_{i}}{\mathrm{~d} t}-
                \sum_{j} m_{i} m_{j}\left(\boldsymbol{v}_{i}-
                \boldsymbol{v}_{j}\right) \cdot\left[\frac{P_{i}}
                {\bar{\rho}_{i}^{2}} f_{i, j} \nabla_{i}
                W_{i j}\left(h_{i}\right)\right],

        it should have been,

            .. math::
                \frac{\mathrm{d} E_{i}}{\mathrm{~d} t}=\boldsymbol{v}_{i}
                \cdot \frac{\mathrm{d} \boldsymbol{P}_{i}}{\mathrm{~d} t}+
                \sum_{j} m_{i} m_{j}\left(\boldsymbol{v}_{i}-
                \boldsymbol{v}_{j}\right) \cdot\left[\frac{P_{i}}
                {\bar{\rho}_{i}^{2}} f_{i, j} \nabla_{i}
                W_{i j}\left(h_{i}\right)\right].

        Specific thermal energy, :math:`u`, would therefore be evolved
        using,

            .. math::
                \frac{\mathrm{d} u_{i}}{\mathrm{~d} t}=
                \sum_{j} m_{j}\left(\boldsymbol{v}_{i}-
                \boldsymbol{v}_{j}\right) \cdot\left[\frac{P_{i}}
                {\bar{\rho}_{i}^{2}} f_{i, j} \nabla_{i}
                W_{i j}\left(h_{i}\right)\right]
        """
        self.beta = beta
        self.dim = dim
        self.fkern = fkern
        super().__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw, d_ae):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0
        d_ae[d_idx] = 0.0

        # d_dt_cfl[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_m, s_m, d_p, s_p, d_cs, s_cs, d_rho, s_rho,
             d_au, d_av, d_aw, d_ae, XIJ, VIJ, DWI, DWJ, HIJ, d_alpha, s_alpha,
             R2IJ, RHOIJ1, d_h, d_dndh, d_n, d_drhosumdh, s_h, s_dndh, s_n,
             s_drhosumdh):
        avi = declare("matrix(3)")
        dim = self.dim

        # particle pressure
        p_i = d_p[d_idx]
        pj = s_p[s_idx]

        # p_i/rhoi**2
        rhoi2 = d_rho[d_idx] * d_rho[d_idx]
        pibrhoi2 = p_i / rhoi2

        # pj/rhoj**2
        rhoj2 = s_rho[s_idx] * s_rho[s_idx]
        pjbrhoj2 = pj / rhoj2

        # averaged sound speed
        cij = 0.5 * (d_cs[d_idx] + s_cs[s_idx])

        mj = s_m[s_idx]
        hij = self.fkern * HIJ
        vijdotxij = (VIJ[0] * XIJ[0] + VIJ[1] * XIJ[1] + VIJ[2] * XIJ[2])

        # Artificial viscosity
        if vijdotxij <= 0.0:
            # viscosity
            alpha = 0.5 * (d_alpha[d_idx] + s_alpha[s_idx])
            muij = hij * vijdotxij / (R2IJ + 0.0001 * hij * hij)
            common = alpha * muij * (cij - self.beta * muij) * mj * RHOIJ1 / 2

            avi[0] = common * (DWI[0] + DWJ[0])
            avi[1] = common * (DWI[1] + DWJ[1])
            avi[2] = common * (DWI[2] + DWJ[2])

            # viscous contribution to velocity
            d_au[d_idx] += avi[0]
            d_av[d_idx] += avi[1]
            d_aw[d_idx] += avi[2]

            # viscous contribution to the thermal energy
            d_ae[d_idx] -= 0.5 * (VIJ[0] * avi[0] +
                                  VIJ[1] * avi[1] +
                                  VIJ[2] * avi[2])

        # grad-h correction terms.
        hibynidim = d_h[d_idx] / (d_n[d_idx] * dim)
        inbrkti = 1 + d_dndh[d_idx] * hibynidim
        inprthsi = d_drhosumdh[d_idx] * hibynidim
        fij = 1 - inprthsi / (s_m[s_idx] * inbrkti)

        hjbynjdim = s_h[s_idx] / (s_n[s_idx] * dim)
        inbrktj = 1 + s_dndh[s_idx] * hjbynjdim
        inprthsj = s_drhosumdh[s_idx] * hjbynjdim
        fji = 1 - inprthsj / (d_m[d_idx] * inbrktj)

        # accelerations for velocity
        comi = mj * pibrhoi2 * fij
        comj = mj * pjbrhoj2 * fji

        d_au[d_idx] -= comi * DWI[0] + comj * DWJ[0]
        d_av[d_idx] -= comi * DWI[1] + comj * DWJ[1]
        d_aw[d_idx] -= comi * DWI[2] + comj * DWJ[2]

        # accelerations for the thermal energy
        vijdotdwi = VIJ[0] * DWI[0] + VIJ[1] * DWI[1] + VIJ[2] * DWI[2]
        d_ae[d_idx] += comi * vijdotdwi


class WallBoundary(Equation):
    """
        :class:`WallBoundary
        <pysph.sph.gas_dynamics.boundary_equations.WallBoundary>` modified
        for TSPH.

        Most importantly, mass of the boundary particle should never be zero
        since it appears in denominator of fij. This has been addressed.
    """

    def initialize(self, d_idx, d_p, d_rho, d_e, d_m, d_cs, d_h, d_htmp, d_h0,
                   d_u, d_v, d_w, d_wij, d_n, d_dndh, d_drhosumdh, d_divv,
                   d_m0):
        d_p[d_idx] = 0.0
        d_u[d_idx] = 0.0
        d_v[d_idx] = 0.0
        d_w[d_idx] = 0.0
        d_m0[d_idx] = d_m[d_idx]
        d_m[d_idx] = 0.0
        d_rho[d_idx] = 0.0
        d_e[d_idx] = 0.0
        d_cs[d_idx] = 0.0
        d_divv[d_idx] = 0.0
        d_wij[d_idx] = 0.0
        d_h[d_idx] = d_h0[d_idx]
        d_htmp[d_idx] = 0.0
        d_n[d_idx] = 0.0
        d_dndh[d_idx] = 0.0
        d_drhosumdh[d_idx] = 0.0

    def loop(self, d_idx, s_idx, d_p, d_rho, d_e, d_m, d_cs, d_divv, d_h, d_u,
             d_v, d_w, d_wij, d_htmp, s_p, s_rho, s_e, s_m, s_cs, s_h, s_divv,
             s_u, s_v, s_w, WI, s_n, d_n, s_dndh, d_dndh, d_drhosumdh,
             s_drhosumdh):
        d_wij[d_idx] += WI
        d_p[d_idx] += s_p[s_idx] * WI
        d_u[d_idx] -= s_u[s_idx] * WI
        d_v[d_idx] -= s_v[s_idx] * WI
        d_w[d_idx] -= s_w[s_idx] * WI
        d_m[d_idx] += s_m[s_idx] * WI
        d_rho[d_idx] += s_rho[s_idx] * WI
        d_e[d_idx] += s_e[s_idx] * WI
        d_cs[d_idx] += s_cs[s_idx] * WI
        d_divv[d_idx] += s_divv[s_idx] * WI
        d_htmp[d_idx] += s_h[s_idx] * WI
        d_n[d_idx] += s_n[s_idx] * WI
        d_dndh[d_idx] += s_dndh[s_idx] * WI
        d_drhosumdh[d_idx] += s_drhosumdh[s_idx] * WI

    def post_loop(self, d_idx, d_p, d_rho, d_e, d_m, d_cs, d_divv, d_h, d_u,
                  d_v, d_w, d_wij, d_htmp, d_n, d_dndh, d_drhosumdh, d_m0):
        if d_wij[d_idx] > 1e-30:
            d_p[d_idx] = d_p[d_idx] / d_wij[d_idx]
            d_u[d_idx] = d_u[d_idx] / d_wij[d_idx]
            d_v[d_idx] = d_v[d_idx] / d_wij[d_idx]
            d_w[d_idx] = d_w[d_idx] / d_wij[d_idx]
            d_m[d_idx] = d_m[d_idx] / d_wij[d_idx]
            d_rho[d_idx] = d_rho[d_idx] / d_wij[d_idx]
            d_e[d_idx] = d_e[d_idx] / d_wij[d_idx]
            d_cs[d_idx] = d_cs[d_idx] / d_wij[d_idx]
            d_divv[d_idx] = d_divv[d_idx] / d_wij[d_idx]
            d_h[d_idx] = d_htmp[d_idx] / d_wij[d_idx]
            d_n[d_idx] = d_n[d_idx] / d_wij[d_idx]
            d_dndh[d_idx] = d_dndh[d_idx] / d_wij[d_idx]
            d_drhosumdh[d_idx] = d_drhosumdh[d_idx] / d_wij[d_idx]

        # Secret Sauce
        if d_m[d_idx] < 1e-10:
            d_m[d_idx] = d_m0[d_idx]


class UpdateGhostProps(Equation):
    def __init__(self, dest, sources=None, dim=2):
        """
        :class:`MPMUpdateGhostProps
        <pysph.sph.gas_dynamics.basic.MPMUpdateGhostProps>` modified
        for TSPH
        """
        super().__init__(dest, sources)
        self.dim = dim
        assert GHOST_TAG == 2

    def initialize(self, d_idx, d_orig_idx, d_p, d_tag, d_h, d_rho, d_dndh,
                   d_psumdh, d_n):
        idx = declare('int')
        if d_tag[d_idx] == 2:
            idx = d_orig_idx[d_idx]
            d_p[d_idx] = d_p[idx]
            d_h[d_idx] = d_h[idx]
            d_rho[d_idx] = d_rho[idx]
            d_dndh[d_idx] = d_dndh[idx]
            d_psumdh[d_idx] = d_psumdh[idx]
            d_n[d_idx] = d_n[idx]


class PECStep(IntegratorStep):
    """Predictor Corrector integrator for Gas-dynamics modified for TSPH"""

    def initialize(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_h, d_u0,
                   d_v0, d_w0, d_u, d_v, d_w, d_e, d_e0, d_h0, d_converged,
                   d_rho, d_rho0, d_n, d_n0):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_e0[d_idx] = d_e[d_idx]

        d_h0[d_idx] = d_h[d_idx]
        d_rho0[d_idx] = d_rho[d_idx]
        d_n0[d_idx] = d_n[d_idx]

        # set the converged attribute to 0 at the beginning of a Group
        d_converged[d_idx] = 0

    def stage1(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_u0, d_v0, d_w0,
               d_u, d_v, d_w, d_e0, d_e, d_au, d_av, d_aw, d_ae, d_rho, d_rho0,
               d_arho, d_h, d_h0, d_ah, dt, d_n, d_n0, d_an):
        dtb2 = 0.5 * dt

        d_u[d_idx] = d_u0[d_idx] + dtb2 * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dtb2 * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dtb2 * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dtb2 * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dtb2 * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dtb2 * d_w[d_idx]

        # update thermal energy
        d_e[d_idx] = d_e0[d_idx] + dtb2 * d_ae[d_idx]

        # predict density and smoothing lengths for faster
        # convergence. NNPS need not be explicitly updated since it
        # will be called at the end of the predictor stage.
        d_h[d_idx] = d_h0[d_idx] + dtb2 * d_ah[d_idx]
        d_rho[d_idx] = d_rho0[d_idx] + dtb2 * d_arho[d_idx]
        d_n[d_idx] = d_n0[d_idx] + dtb2 * d_an[d_idx]

    def stage2(self, d_idx, d_x0, d_y0, d_z0, d_x, d_y, d_z, d_u0, d_v0, d_w0,
               d_u, d_v, d_w, d_e0, d_e, d_au, d_av, d_aw, d_ae, dt):
        d_u[d_idx] = d_u0[d_idx] + dt * d_au[d_idx]
        d_v[d_idx] = d_v0[d_idx] + dt * d_av[d_idx]
        d_w[d_idx] = d_w0[d_idx] + dt * d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + dt * d_u[d_idx]
        d_y[d_idx] = d_y0[d_idx] + dt * d_v[d_idx]
        d_z[d_idx] = d_z0[d_idx] + dt * d_w[d_idx]

        # Update densities and smoothing lengths from the accelerations
        d_e[d_idx] = d_e0[d_idx] + dt * d_ae[d_idx]



"""Utility functions to read Daniel Price's NDSPMHD solution files"""

import struct
from pysph.base.utils import get_particle_array_gasd as gpa

from .fortranfile import FortranFile

def ndspmhd2pysph(fname, dim=2, read_type=False):
    """Read output data file from NDSPMHD

    Parameters:
    
    fname : str
        NDSPMHD data filename

    dim : int
        Problem dimension

    read_type : bint
        Flag to read the `type` property for particles

    Returns the ParticleArray representation of the data that can be
    used in PySPH.

    """
    f = FortranFile(fname)

    # get the header length
    header_length = f._header_length
    endian = f.ENDIAN
    
    # get the length of the record to be read
    length = f._read_check()

    # now read the individual entries:

    # current time : double
    t = f._read_exactly(8)
    t = struct.unpack(endian+"1d", t)[0]

    # number of particles and number printed : int
    npart = f._read_exactly(4)
    nprint = f._read_exactly(4)

    npart = struct.unpack(endian+"1i", npart)[0]
    nprint = struct.unpack(endian+"1i", nprint)[0]

    # gamma and hfact : double
    gamma = f._read_exactly(8)
    hfact = f._read_exactly(8)

    gamma = struct.unpack(endian+"1d", gamma)[0]
    hfact = struct.unpack(endian+"1d", hfact)[0]

    # ndim, ndimV : int
    ndim = f._read_exactly(4)
    ndimV = f._read_exactly(4)

    # ncollumns, iformat, ibound : int
    nc = f._read_exactly(4)
    ifmt = f._read_exactly(4)
    ib1 = f._read_exactly(4)
    ib2 = f._read_exactly(4)

    nc = struct.unpack(endian+"1i", nc)[0]

    # xmin, xmax : double
    xmin1 = f._read_exactly(8)
    xmin2 = f._read_exactly(8)
    xmax1 = f._read_exactly(8)
    xmax2 = f._read_exactly(8)

    # n : int
    n = f._read_exactly(4)
    n = struct.unpack(endian+"1i", n)[0]

    # geometry type
    geom = f._read_exactly(n)

    # end reading this header
    f._read_check()

    # Now go on to the arrays. Remember, there are 16 entries
    # correcponding to the columns

    x = f.readReals(prec="d")
    y = f.readReals(prec="d")
    u = f.readReals(prec="d")
    v = f.readReals(prec="d")
    w = f.readReals(prec="d")

    h = f.readReals(prec="d")
    rho = f.readReals(prec="d")
    e = f.readReals(prec="d")
    m = f.readReals(prec="d")

    alpha1 = f.readReals(prec="d")
    alpha2 = f.readReals(prec="d")

    p = f.readReals(prec="d")
    drhobdtbrho = f.readReals("d")
    gradh = f.readReals("d")

    au = f.readReals("d")
    av = f.readReals("d")
    aw = f.readReals("d")

    # By default, NDSPMHD does not output the type array. You need to
    # add this to the output routine if you want it.
    if read_type:
        type = f.readInts(prec="i")

    # now create the particle array
    pa = gpa(name='fluid', x=x, y=y, m=m, h=h, rho=rho, e=e, p=p, 
             u=u, v=v, w=w, au=au, av=av, aw=aw, div=drhobdtbrho)

    return pa



"""Two-dimensional dam break over a dry bed.  (30 minutes)

The case is described in "State of the art classical SPH for free surface
flows", Moncho Gomez-Gesteira, Benedict D Rogers, Robert A, Dalrymple and Alex
J.C Crespo, Journal of Hydraulic Research, Vol 48, Extra Issue (2010), pp
6-27. DOI:10.1080/00221686.2010.9641242

"""

import os
import numpy as np

from pysph.base.kernels import WendlandQuintic, QuinticSpline
from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme, SchemeChooser, AdamiHuAdamsScheme
from pysph.sph.wc.edac import EDACScheme
from pysph.sph.iisph import IISPHScheme
from pysph.sph.equation import Group
from pysph.sph.wc.kernel_correction import (GradientCorrectionPreStep,
                                            GradientCorrection,
                                            MixedKernelCorrectionPreStep)
from pysph.sph.wc.crksph import CRKSPHPreStep, CRKSPH
from pysph.sph.wc.gtvf import GTVFScheme
from pysph.sph.isph.sisph import SISPHScheme
from pysph.tools.geometry import get_2d_tank, get_2d_block


fluid_column_height = 2.0
fluid_column_width = 1.0
container_height = 4.0
container_width = 4.0
nboundary_layers = 2
nu = 0.0
dx = 0.03
g = 9.81
ro = 1000.0
vref = np.sqrt(2 * 9.81 * fluid_column_height)
co = 10.0 * vref
gamma = 7.0
alpha = 0.1
beta = 0.0
B = co * co * ro / gamma
p0 = 1000.0
hdx = 1.3
h = hdx * dx
m = dx**2 * ro


class DamBreak2D(Application):

    def add_user_options(self, group):
        corrections = ['', 'mixed-corr', 'grad-corr', 'kernel-corr', 'crksph']
        group.add_argument(
            '--dx', action='store', type=float, dest='dx', default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx', default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )
        group.add_argument(
            "--kernel-corr", action="store", type=str, dest='kernel_corr',
            default='', help="Type of Kernel Correction", choices=corrections
        )
        group.add_argument(
            '--staggered-grid', action="store_true", dest='staggered_grid',
            default=False, help="Use a staggered grid for particles.",
        )

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h = self.hdx * self.dx
        self.kernel_corr = self.options.kernel_corr
        self.co = co
        print("Using h = %f" % self.h)

    def configure_scheme(self):
        tf = 2.5
        kw = dict(
            tf=tf, output_at_times=[0.4, 0.6, 0.8, 1.0]
        )
        co = self.co
        B = co * co * ro / gamma
        if self.options.scheme == 'wcsph':
            dt = 0.125 * self.h / co
            self.scheme.configure(h0=self.h, hdx=self.hdx)
            kernel = WendlandQuintic(dim=2)
            from pysph.sph.integrator import PECIntegrator
            kw.update(
                dict(
                    integrator_cls=PECIntegrator,
                    kernel=kernel, adaptive_timestep=True, n_damp=50,
                    fixed_h=False, dt=dt
                )
            )
        elif self.options.scheme == 'aha':
            self.scheme.configure(h0=self.h)
            kernel = QuinticSpline(dim=2)
            dt = 0.125 * self.h / co
            kw.update(
                dict(
                    kernel=kernel, dt=dt
                )
            )
            print("dt = %f" % dt)
        elif self.options.scheme == 'edac':
            self.scheme.configure(h=self.h)
            kernel = QuinticSpline(dim=2)
            dt = 0.125 * self.h / co
            kw.update(
                dict(
                    kernel=kernel, dt=dt
                )
            )
            print("dt = %f" % dt)
        elif self.options.scheme == 'iisph':
            kernel = QuinticSpline(dim=2)
            dt = 0.125 * 10 * self.h / co
            kw.update(
                dict(
                    kernel=kernel, dt=dt, adaptive_timestep=True
                )
            )
            print("dt = %f" % dt)
        elif self.options.scheme == 'gtvf':
            scheme = self.scheme
            kernel = QuinticSpline(dim=2)
            dt = 0.125 * self.h / co
            kw.update(dict(kernel=kernel, dt=dt))
            scheme.configure(pref=B*gamma, h0=self.h)
            print("dt = %f" % dt)
        elif self.options.scheme == 'sisph':
            dt = 0.125*self.h/vref
            kernel = QuinticSpline(dim=2)
            print("SISPH dt = %f" % dt)
            kw.update(dict(kernel=kernel))
            self.scheme.configure_solver(
                dt=dt, tf=tf, adaptive_timestep=False, pfreq=10,
            )

        self.scheme.configure_solver(**kw)

    def create_scheme(self):
        wcsph = WCSPHScheme(
            ['fluid'], ['boundary'], dim=2, rho0=ro, c0=co,
            h0=h, hdx=1.3, gy=-9.81, alpha=alpha, beta=beta,
            gamma=gamma, hg_correction=True, update_h=True
        )
        aha = AdamiHuAdamsScheme(
            fluids=['fluid'], solids=['boundary'], dim=2, c0=co, nu=nu,
            rho0=ro, h0=h, p0=0.0, gy=-g, gamma=1.0, tdamp=0.0, alpha=alpha
        )
        edac = EDACScheme(
            fluids=['fluid'], solids=['boundary'], dim=2, c0=co, nu=nu,
            rho0=ro, h=h, pb=0.0, gy=-g, eps=0.0, clamp_p=True
        )
        iisph = IISPHScheme(
            fluids=['fluid'], solids=['boundary'], dim=2, nu=nu,
            rho0=ro, gy=-g
        )
        gtvf = GTVFScheme(
            fluids=['fluid'], solids=['boundary'], dim=2, nu=nu,
            rho0=ro, gy=-g, h0=None, c0=co, pref=None
        )
        sisph = SISPHScheme(
            fluids=['fluid'], solids=['boundary'], dim=2, nu=nu,
            c0=co, rho0=ro, alpha=0.05, gy=-g, pref=ro*co**2,
            internal_flow=False, hg_correction=True, gtvf=True, symmetric=True
        )
        s = SchemeChooser(default='wcsph', wcsph=wcsph, aha=aha, edac=edac,
                          iisph=iisph, gtvf=gtvf, sisph=sisph)
        return s

    def create_equations(self):
        eqns = self.scheme.get_equations()
        if self.options.scheme == 'iisph' or self.options.scheme == 'sisph':
            return eqns
        if self.options.scheme == 'gtvf':
            return eqns
        n = len(eqns)
        if self.kernel_corr == 'grad-corr':
            eqn1 = Group(equations=[
                GradientCorrectionPreStep('fluid', ['fluid', 'boundary'])
            ], real=False)
            for i in range(n):
                eqn2 = GradientCorrection('fluid', ['fluid', 'boundary'])
                eqns[i].equations.insert(0, eqn2)
            eqns.insert(0, eqn1)
        elif self.kernel_corr == 'mixed-corr':
            eqn1 = Group(equations=[
                MixedKernelCorrectionPreStep('fluid', ['fluid', 'boundary'])
            ], real=False)
            for i in range(n):
                eqn2 = GradientCorrection('fluid', ['fluid', 'boundary'])
                eqns[i].equations.insert(0, eqn2)
            eqns.insert(0, eqn1)
        elif self.kernel_corr == 'crksph':
            eqn1 = Group(equations=[
                CRKSPHPreStep('fluid', ['fluid', 'boundary']),
                CRKSPHPreStep('boundary', ['fluid', 'boundary'])
            ], real=False)
            for i in range(n):
                eqn2 = CRKSPH('fluid', ['fluid', 'boundary'])
                eqn3 = CRKSPH('boundary', ['fluid', 'boundary'])
                eqns[i].equations.insert(0, eqn3)
                eqns[i].equations.insert(0, eqn2)
            eqns.insert(0, eqn1)
        return eqns

    def create_particles(self):
        if self.options.staggered_grid:
            nboundary_layers = 2
            nfluid_offset = 2
            wall_hex_pack = True
        else:
            nboundary_layers = 4
            nfluid_offset = 1
            wall_hex_pack = False
        xt, yt = get_2d_tank(dx=self.dx, length=container_width,
                             height=container_height, base_center=[2, 0],
                             num_layers=nboundary_layers)
        xf, yf = get_2d_block(dx=self.dx, length=fluid_column_width,
                              height=fluid_column_height, center=[0.5, 1])

        xf += self.dx
        yf += self.dx

        fluid = get_particle_array(name='fluid', x=xf, y=yf, h=h, m=m, rho=ro)
        boundary = get_particle_array(name='boundary', x=xt, y=yt, h=h, m=m,
                                      rho=ro)

        self.scheme.setup_properties([fluid, boundary])
        if self.options.scheme == 'iisph':
            # the default position tends to cause the particles to be pushed
            # away from the wall, so displacing it by a tiny amount helps.
            fluid.x += self.dx / 4

        self.kernel_corrections(fluid, boundary)
        return [fluid, boundary]

    def kernel_corrections(self, fluid, boundary):
        # Adding extra properties for kernel correction
        corr = self.kernel_corr
        if corr == 'kernel-corr' or corr == 'mixed-corr':
            fluid.add_property('cwij')
            boundary.add_property('cwij')
        if corr == 'mixed-corr' or corr == 'grad-corr':
            fluid.add_property('m_mat', stride=9)
            boundary.add_property('m_mat', stride=9)
        elif corr == 'crksph':
            fluid.add_property('ai')
            boundary.add_property('ai')
            fluid.add_property('gradbi', stride=9)
            boundary.add_property('gradbi', stride=9)
            for prop in ['gradai', 'bi']:
                fluid.add_property(prop, stride=3)
                boundary.add_property(prop, stride=3)

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        t, x_max = [], []
        factor = np.sqrt(2.0 * 9.81 / fluid_column_width)
        files = self.output_files
        for sd, array in iter_output(files, 'fluid'):
            t.append(sd['t'] * factor)
            x = array.get('x')
            x_max.append(x.max())

        t, x_max = list(map(np.asarray, (t, x_max)))
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, t=t, x_max=x_max)

        import matplotlib
        matplotlib.use('Agg')
        from matplotlib import pyplot as plt
        from pysph.examples import db_exp_data as dbd
        plt.plot(t, x_max, label='Computed')
        te, xe = dbd.get_koshizuka_oka_data()
        plt.plot(te, xe, 'o', label='Koshizuka & Oka (1996)')
        plt.xlim(0, 0.7 * factor)
        plt.ylim(0, 4.5)
        plt.xlabel('$T$')
        plt.ylabel('$Z/L$')
        plt.legend(loc='upper left')
        plt.savefig(os.path.join(self.output_dir, 'x_vs_t.png'), dpi=300)
        plt.close()

    def customize_output(self):
        self._mayavi_config('''
        b = particle_arrays['fluid']
        b.scalar = 'vmag'
        ''')


if __name__ == '__main__':
    app = DamBreak2D()
    app.run()
    app.post_process(app.info_filename)



r"""Diffusion of an acoustic wave in 1-d (5 minutes)

Propagation of acoustic wave
particles have properties according
to the following distribuion
.. math::
        \rho = \rho_0 + \Delta\rho sin(kx)
        p = p_0 + c_0^2\Delta\rho sin(kx)
        u = c_0\rho_0^{-1}\Delta\rho sin(kx)

with :math:`\Delta\rho = 1e-6` and :math:`k = 2\pi/\lambda`
where :math:`\lambda` is the domain length.
.. math::
        \rho_0 = \gamma = 1.4 and p_0 = 1.0
"""


# standard library and numpy imports
import numpy

# pysph imports
from pysph.base.utils import get_particle_array as gpa
from pysph.base.nnps import DomainManager
from pysph.solver.application import Application
from pysph.sph.scheme import \
    GSPHScheme, ADKEScheme, GasDScheme, SchemeChooser
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme


class AcousticWave(Application):
    def initialize(self):
        self.xmin = 0.
        self.xmax = 1.
        self.gamma = 1.4
        self.rho_0 = self.gamma
        self.p_0 = 1.
        self.c_0 = 1.
        self.delta_rho = 1e-6
        self.n_particles = 8
        self.domain_length = self.xmax - self.xmin
        self.k = -2 * numpy.pi / self.domain_length
        self.cfl = 0.1
        self.hdx = 1.0
        self.dt = 1e-3
        self.tf = 5
        self.dim = 1

    def create_domain(self):
        return DomainManager(
            xmin=0, xmax=1, periodic_in_x=True
        )

    def add_user_options(self, group):
        group.add_argument(
            "--nparticles", action="store", type=int, dest="nprt", default=256,
            help="Number of particles in domain"
        )

    def consume_user_options(self):
        self.n_particles = self.options.nprt
        self.dx = self.domain_length / (self.n_particles)
        self.dt = self.cfl * self.dx / self.c_0

    def create_particles(self):
        x = numpy.arange(
            self.xmin + self.dx*0.5, self.xmax, self.dx
        )
        rho = self.rho_0 + self.delta_rho *\
            numpy.sin(self.k * x)

        p = self.p_0 + self.c_0**2 *\
            self.delta_rho * numpy.sin(self.k * x)

        u = self.c_0 * self.delta_rho * numpy.sin(self.k * x) /\
            self.rho_0
        cs = numpy.sqrt(
            self.gamma * p / rho
        )
        h = numpy.ones_like(x) * self.dx * self.hdx
        m = numpy.ones_like(x) * self.dx * rho
        e = p / ((self.gamma - 1) * rho)
        fluid = gpa(
            name='fluid', x=x, p=p, rho=rho, u=u, h=h, m=m, e=e, cs=cs,
            h0=h.copy()
        )
        self.scheme.setup_properties([fluid])

        return [fluid, ]

    def create_scheme(self):
        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim,
            gamma=self.gamma, kernel_factor=1.0,
            g1=0., g2=0., rsolver=7, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=5.0,
            niter=40, tol=1e-6, has_ghosts=True
        )

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            kernel_factor=1.2, alpha1=0, alpha2=0,
            beta=2.0, update_alpha1=False, update_alpha2=False,
            has_ghosts=True
        )

        crksph = CRKSPHScheme(
            fluids=['fluid'], dim=self.dim, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=self.gamma, cl=2, has_ghosts=True
        )

        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            alpha=0, beta=0.0, k=1.5, eps=0.0, g1=0.0, g2=0.0,
            has_ghosts=True)

        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            hfact=1.2
        )

        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=self.dim, gamma=self.gamma,
            hfact=1.2)

        s = SchemeChooser(
            default='gsph', gsph=gsph, mpm=mpm, crksph=crksph, adke=adke,
            psph=psph, tsph=tsph
        )

        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'gsph':
            s.configure_solver(
                dt=self.dt, tf=self.tf,
                adaptive_timestep=True, pfreq=50
            )

        elif self.options.scheme == 'mpm':
            s.configure(kernel_factor=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

        elif self.options.scheme == 'crksph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

    def post_process(self):
        import os
        from pysph.solver.utils import load
        if len(self.output_files) < 1:
            return
        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']
        x_c = pa.x
        u = self.c_0 * self.delta_rho * numpy.sin(self.k * x_c) /\
            self.rho_0
        u_c = pa.u
        l_inf = numpy.max(
            numpy.abs(u_c - u)
        )
        l_1 = (numpy.sum(
            numpy.abs(u_c - u)
        ) / self.n_particles)
        print("L_inf norm of velocity for the problem: %s" % (l_inf))
        print("L_1 norm of velocity for the problem: %s" % (l_1))

        rho = self.rho_0 + self.delta_rho *\
            numpy.sin(self.k * x_c)

        rho_c = pa.rho
        l1 = numpy.sum(
            numpy.abs(rho - rho_c)
        )
        l1 = l1 / self.n_particles
        print("l_1 norm of density for the problem: %s" % (l1))
        fname = os.path.join(self.output_dir, 'norms.npz')
        numpy.savez(fname, linf_vel=l_inf, l1_vel=l_1, l1_rho=l1)


if __name__ == "__main__":
    app = AcousticWave()
    app.run()
    app.post_process()



'''This helper module orchestrates the generation of OpenCL/CUDA code, compiles
it and makes it available for use.

Overview
~~~~~~~~~

Look first at sph/tests/test_acceleration_eval.py to see the big picture. The
general idea when using AccelerationEval instances is:

- Create the particle arrays.
- Specify any equations and the SPH kernel.
- Construct the AccelerationEval with the particles, equations and kernel.
- Compile this with SPHCompiler and hand in an NNPS.
  - For the GPU all that changes is the backend and the NNPS.

So the difference in the CPU version and GPU version is the choice of the
backend. The AccelerationEval delegates its actual high-performance work to its
`self.c_acceleration_eval` instance. This instance is either compiled with
Cython or OpenCL. With Cython this is actually a compiled extension module
created with Cython and with OpenCL this is the Python class
OpenCLAccelerationEval in this file. This is where the helpers come in.


The AccelerationEvalCythonHelper and AccelerationEvalOpenCLHelper have three
main methods:

- get_code(): returns the code to be compiled.

- compile(code): compile the code and return the compiled module/opencl
  Program.

- setup_compiled_module(module): sets the AccelerationEval's
  c_acceleration_eval to an instance based on the helper.

The helper basically uses mako templates, code generation via simple string
manipulations, and transpilation to generate HPC code automatically from the
given particle arrays, equations, and kernel.

In this module, an OpenCLAccelerationEval is defined which does the work of
calling the compiled opencl kernels. The AccelerationEvalOpenCLHelper generates
the OpenCL kernels. The general idea of how we generate OpenCL kernels is quite
simple.

We transpile pure Python code using `pysph.base.translator` which generates C
from a subset of pure Python.

- We do not support inheritance but convert classes to simple C-structs and
  functions which take the struct as the first argument.
- Python functions are also transpiled.
- Type inference is done using either conventions like s_idx, d_idx, s_x, d_x,
  WIJ etc. or by type hints given using default arguments. Lists are treated as
  raw pointers to the contained type. One can also set certain predefined
  known_types and the code generator will generate suitable code. There are
  plenty of tests illustrating what is supported in
  ``pysph.base.tests.test_translator``.
- One can also use the ``declare`` function to declare any types in the Python
  code.

This is enough to do what we need. We transpile the kernel, all required
equations, and generate suitable kernels. All structs are converted to suitable
GPU types and the data from the Python classes is converted into suitably
aligned numpy dtypes (using cl.tools.match_dtype_to_c_struct). These are
constructed for each class and stored in an _gpu attribute on the Python
object.  When calling kernels these are passed and pushed/pulled from the GPU.

When the user calls AccelerationEval.compute, this in turn calls the
c_acceleration_eval's compute method. For OpenCL, this is provided by the
OpenCLAccelerationEval class below.

While the implementation is a bit complex, the details a bit hairy, the general
idea is very simple.

'''
from functools import partial
import inspect
import logging
import os
import re
import sys
from textwrap import wrap

import numpy as np
from mako.template import Template

from pysph.base.utils import is_overloaded_method
from pysph.base.device_helper import DeviceHelper

from pysph.sph.acceleration_nnps_helper import generate_body, \
    get_kernel_args_list

from compyle.ext_module import get_platform_dir, get_md5
from compyle.config import get_config
from compyle.profile import profile_ctx
from compyle.translator import (CStructHelper, CUDAConverter, OpenCLConverter,
                                ocl_detect_type, ocl_detect_pointer_base_type)

from .equation import get_predefined_types, KnownType
from .acceleration_eval_cython_helper import (
    get_all_array_names, get_known_types_for_arrays
)

logger = logging.getLogger(__name__)

getfullargspec = getattr(
    inspect, 'getfullargspec', inspect.getargspec
)


def get_converter(backend):
    if backend == 'opencl':
        Converter = OpenCLConverter
    elif backend == 'cuda':
        Converter = CUDAConverter
    else:
        raise RuntimeError('Invalid backend: %s' % backend)
    return Converter


def get_kernel_definition(kernel, arg_list):
    sig = 'KERNEL void\n{kernel}\n({args})'.format(
        kernel=kernel, args=', '.join(arg_list),
    )
    return '\n'.join(wrap(sig, width=78, subsequent_indent=' ' * 4,
                          break_long_words=False))


def wrap_code(code, indent=' ' * 4):
    return wrap(
        code, width=74, initial_indent=indent,
        subsequent_indent=indent + ' ' * 4, break_long_words=False
    )


def get_helper_code(helpers, transpiler=None, backend=None):
    """This function generates any additional code for the given list of
    helpers.
    """
    result = []
    if transpiler is None:
        transpiler = get_converter(backend)
    doc = '\n// Helpers.\n'
    result.append(doc)
    for helper in helpers:
        result.append(transpiler.parse_function(helper))
    return result


class DummyQueue(object):
    def finish(self):
        pass


def get_context(backend):
    if backend == 'cuda':
        from compyle.cuda import set_context
        set_context()
        from pycuda.autoinit import context
        return context
    elif backend == 'opencl':
        from compyle.opencl import get_context
        return get_context()
    else:
        raise RuntimeError('Unsupported GPU backend %s' % backend)


def get_queue(backend):
    if backend == 'cuda':
        return DummyQueue()
    elif backend == 'opencl':
        from compyle.opencl import get_queue
        return get_queue()
    else:
        raise RuntimeError('Unsupported GPU backend %s' % backend)


def profile_kernel(knl, backend):
    if backend == 'cuda' or backend == 'opencl':
        from compyle.profile import profile_kernel
        return profile_kernel(knl, knl.function_name, backend=backend)
    else:
        raise RuntimeError('Unsupported GPU backend %s' % backend)


class GPUAccelerationEval(object):
    """Does the actual work of performing the evaluation.
    """

    def __init__(self, helper):
        self.helper = helper
        self.particle_arrays = helper.object.particle_arrays
        self.nnps = None
        self._queue = helper._queue
        cfg = get_config()
        self._use_double = cfg.use_double
        self._use_local_memory = cfg.use_local_memory

    def _get_index(self, dest, attr):
        if isinstance(attr, str):
            return dest.get(attr)[0]
        else:
            return attr

    def _call_kernel(self, info, extra_args):
        nnps = self.nnps
        call = info.get('method')
        args = list(info.get('args'))
        dest = info['dest']
        start_idx = self._get_index(dest, info['start_idx'])
        stop_idx = info['stop_idx']
        if stop_idx is None:
            n = dest.get_number_of_particles(info.get('real', True))
        else:
            n = self._get_index(dest, stop_idx)
        args[1] = (n - start_idx,)
        args[3:] = [x() for x in args[3:]]
        # Argument for NP_MAX
        extra_args[-2][...] = n - 1
        # Argument for OFFSET_IDX
        extra_args[-1][...] = start_idx

        if info.get('loop'):
            if self._use_local_memory:
                nnps.set_context(info['src_idx'], info['dst_idx'])

                nnps_args, gs_ls = self.nnps.get_kernel_args('float')
                self._queue.finish()
                args[1] = gs_ls[0]
                args[2] = gs_ls[1]

                # No need for the guard variable for the local memory code.
                args = args + extra_args[:-2] + nnps_args

                call(*args)
                self._queue.finish()
            else:
                nnps.set_context(info['src_idx'], info['dst_idx'])
                cache = nnps.current_cache
                cache.get_neighbors_gpu()
                self._queue.finish()
                args = args + [
                    cache._nbr_lengths_gpu.dev.data,
                    cache._start_idx_gpu.dev.data,
                    cache._neighbors_gpu.dev.data
                ] + extra_args
                call(*args)
        else:
            call(*(args + extra_args))
        self._queue.finish()

    def _sync_from_gpu(self, eq):
        ary = eq._gpu.get()
        for i, name in enumerate(ary.dtype.names):
            setattr(eq, name, ary[0][i])

    def _converged(self, equations):
        for eq in equations:
            if not (eq.converged() > 0):
                return False
        return True

    def compute(self, t, dt):
        helper = self.helper
        dtype = np.float64 if self._use_double else np.float32
        extra_args = [np.asarray(t, dtype=dtype),
                      np.asarray(dt, dtype=dtype),
                      np.asarray(0, dtype=np.uint32),
                      np.asarray(0, dtype=np.uint32)]
        i = 0
        iter_count = 0
        iter_start = 0
        while i < len(helper.calls):
            info = helper.calls[i]
            type = info['type']
            if type == 'method':
                method_name = info.get('method')
                method = getattr(self, method_name)
                if method_name == 'do_reduce':
                    _args = info.get('args')
                    method(_args[0], _args[1], t, dt)
                else:
                    method(*info.get('args'))
            elif type == 'py_initialize':
                args = info['dest'], t, dt
                for call in info['calls']:
                    call(*args)
            elif type == 'pre_post':
                func = info.get('callable')
                func(*info.get('args'))
            elif type == 'kernel':
                self._call_kernel(info, extra_args)
            elif type == 'check_condition':
                group = info['group']
                if not group.condition(t, dt):
                    # Condition failed so skip group.
                    i = info['jump']
            elif type == 'start_iteration':
                iter_count = 0
                iter_start = i
            elif type == 'stop_iteration':
                eqs = info['equations']
                group = info['group']
                iter_count += 1
                if ((iter_count >= group.min_iterations) and
                        (iter_count == group.max_iterations or
                         self._converged(eqs))):
                    pass
                else:
                    i = iter_start
            i += 1

    def set_nnps(self, nnps):
        self.nnps = nnps

    def update_particle_arrays(self, arrays):
        raise NotImplementedError('GPU backend is incomplete')

    def update_nnps(self):
        with profile_ctx('Integrator.update_domain'):
            self.nnps.update_domain()
        with profile_ctx('nnps.update'):
            self.nnps.update()

    def do_reduce(self, eqs, dest, t, dt):
        for eq in eqs:
            eq.reduce(dest, t, dt)


class CUDAAccelerationEval(GPUAccelerationEval):
    def _call_kernel(self, info, extra_args):
        from pycuda.gpuarray import splay
        import pycuda.driver as drv
        nnps = self.nnps
        call = info.get('method')
        args = list(info.get('args'))
        dest = info['dest']
        start_idx = self._get_index(dest, info['start_idx'])
        stop_idx = info['stop_idx']
        if stop_idx is None:
            n = dest.get_number_of_particles(info.get('real', True))
        else:
            n = self._get_index(dest, stop_idx)
        n_iter = int(n - start_idx)

        # args is actually [queue, None, None, actual_meaningful_args]
        # we do not need the first 3 args on CUDA.
        args = [x() for x in args[3:]]

        # Argument for NP_MAX
        extra_args[-2][...] = n - 1
        # Argument for OFFSET_IDX
        extra_args[-1][...] = start_idx

        gs, ls = splay(n_iter)
        gs, ls = int(gs[0]), int(ls[0])
        num_blocks = (n_iter + ls - 1) // ls

        #num_blocks = int((gs + ls - 1) / ls)
        num_tpb = ls

        if info.get('loop'):
            if self._use_local_memory:
                # FIXME: Fix local memory for CUDA
                nnps.set_context(info['src_idx'], info['dst_idx'])

                nnps_args, gs_ls = self.nnps.get_kernel_args('float')
                args[1] = gs_ls[0]
                args[2] = gs_ls[1]

                # No need for the guard variable for the local memory code.
                args = args + extra_args[:-2] + nnps_args

                call(*args)
            else:
                # find block sizes
                nnps.set_context(info['src_idx'], info['dst_idx'])
                cache = nnps.current_cache
                cache.get_neighbors_gpu()
                args = args + [
                    cache._nbr_lengths_gpu.dev,
                    cache._start_idx_gpu.dev,
                    cache._neighbors_gpu.dev
                ] + extra_args
                event = drv.Event()
                call(*args, block=(num_tpb, 1, 1), grid=(num_blocks, 1))
                event.record()
                event.synchronize()
        else:
            event = drv.Event()
            call(*(args + extra_args),
                 block=(num_tpb, 1, 1),
                 grid=(num_blocks, 1))
            event.record()
            event.synchronize()


def add_address_space(known_types):
    for v in known_types.values():
        if 'GLOBAL_MEM' not in v.type:
            v.type = 'GLOBAL_MEM ' + v.type


def get_equations_with_converged(group):
    def _get_eqs(g):
        if g.has_subgroups:
            res = []
            for x in g.equations:
                res.extend(_get_eqs(x))
            return res
        else:
            return g.equations

    eqs = [x for x in _get_eqs(group)
           if is_overloaded_method(getattr(x, 'converged'))]
    return eqs


def convert_to_float_if_needed(code):
    use_double = get_config().use_double
    if not use_double:
        code = re.sub(r'\bdouble\b', 'float', code)
    return code


class AccelerationEvalGPUHelper(object):
    def __init__(self, acceleration_eval):
        self.object = acceleration_eval
        self.backend = acceleration_eval.backend
        self.all_array_names = get_all_array_names(
            self.object.particle_arrays
        )
        self.known_types = get_known_types_for_arrays(
            self.all_array_names
        )
        add_address_space(self.known_types)
        predefined = dict(get_predefined_types(
            self.object.all_group.pre_comp
        ))
        self.known_types.update(predefined)
        self.known_types['NBRS'] = KnownType('GLOBAL_MEM unsigned int*')
        self.data = []
        self._array_map = None
        self._array_index = None
        self._equations = {}
        self._cpu_structs = {}
        self._gpu_structs = {}
        self.calls = []
        self.program = None
        self._ctx = get_context(self.backend)
        self._queue = get_queue(self.backend)

    def _setup_arrays_on_device(self):
        pas = self.object.particle_arrays
        array_map = {}
        array_index = {}
        for idx, pa in enumerate(pas):
            if pa.gpu is None:
                pa.set_device_helper(DeviceHelper(pa, backend=self.backend))
            array_map[pa.name] = pa
            array_index[pa.name] = idx

        self._array_map = array_map
        self._array_index = array_index

        self._setup_structs_on_device()

    def _setup_structs_on_device(self):
        if self.backend == 'opencl':
            import pyopencl as cl
            import pyopencl.array  # noqa: 401
            import pyopencl.tools  # noqa: 401

            gpu = self._gpu_structs
            cpu = self._cpu_structs
            for k, v in cpu.items():
                if v is None:
                    gpu[k] = v
                else:
                    g_struct, code = cl.tools.match_dtype_to_c_struct(
                        self._ctx.devices[0], "dummy", v.dtype
                    )
                    g_v = v.astype(g_struct)
                    gpu[k] = cl.array.to_device(self._queue, g_v)
                    if k in self._equations:
                        self._equations[k]._gpu = gpu[k]
        else:
            from pycuda import gpuarray
            from compyle.cuda import match_dtype_to_c_struct

            gpu = self._gpu_structs
            cpu = self._cpu_structs
            for k, v in cpu.items():
                if v is None:
                    gpu[k] = v
                else:
                    g_struct, code = match_dtype_to_c_struct(
                        None, "junk", v.dtype
                    )
                    g_v = v.astype(g_struct)
                    gpu[k] = gpuarray.to_gpu(g_v)
                    if k in self._equations:
                        self._equations[k]._gpu = gpu[k]

    def _get_argument(self, arg, dest, src=None):
        ary_map = self._array_map
        structs = self._gpu_structs

        # This is needed for late binding on the device helper's attributes
        # which may change at each iteration when particles are added/removed.
        if self.backend == 'opencl':
            def _get_array(gpu_helper, attr):
                return getattr(gpu_helper, attr).dev.data
        else:
            def _get_array(gpu_helper, attr):
                return getattr(gpu_helper, attr).dev

        def _get_struct(obj):
            return obj

        if arg.startswith('d_'):
            return partial(_get_array, ary_map[dest].gpu, arg[2:])
        elif arg.startswith('s_'):
            return partial(_get_array, ary_map[src].gpu, arg[2:])
        else:
            if self.backend == 'opencl':
                return partial(_get_struct, structs[arg].data)
            else:
                return partial(_get_struct, structs[arg])

    def _setup_calls(self):
        calls = []
        prg = self.program
        array_index = self._array_index
        # Track the condition and end_group to facilitate group conditions.
        condition_stack = []
        count = 0
        for item in self.data:
            type = item.get('type')
            info = {}
            if type == 'kernel':
                kernel = item.get('kernel')
                method = getattr(prg, kernel)
                method = profile_kernel(method, self.backend)
                dest = item['dest']
                src = item.get('source', dest)
                args = [self._queue, None, None]
                for arg in item['args']:
                    args.append(self._get_argument(arg, dest, src))
                loop = item['loop']
                args.append(self._get_argument('kern', dest, src))
                info = dict(
                    method=method, dest=self._array_map[dest],
                    src=self._array_map[src], args=args,
                    loop=loop, src_idx=array_index[src],
                    dst_idx=array_index[dest],
                    start_idx=item.get('start_idx'),
                    stop_idx=item.get('stop_idx'),
                    type='kernel'
                )
            elif type == 'method':
                info = dict(item)
                if info.get('method') == 'do_reduce':
                    args = info.get('args')
                    grp = args[0]
                    args[0] = [x for x in grp.equations
                               if hasattr(x, 'reduce')]
                    args[1] = self._array_map[args[1]]
            elif type == 'pre_post':
                info = dict(item)
            elif type == 'py_initialize':
                info = dict(item)
                info['dest'] = self._array_map[item.get('dest')]
            elif 'iteration' in type:
                group = item['group']
                equations = get_equations_with_converged(group._orig_group)
                info = dict(type=type, equations=equations, group=group)
            elif type == 'check_condition':
                info = dict(item)
                info['jump'] = None
                condition_stack.append(info)
            elif type == 'end_group':
                group = item['group']
                if group.condition is not None:
                    cond_info = condition_stack.pop()
                    # count is the location of the next call.
                    # We decrement it here as the counter is incremented
                    # in the helper.
                    cond_info['jump'] = count - 1
                info = {}
            else:
                raise RuntimeError('Unknown type %s' % type)
            if info:
                calls.append(info)
                count += 1
        return calls

    ##########################################################################
    # Public interface.
    ##########################################################################
    def get_code(self):
        path = os.path.join(os.path.dirname(__file__),
                            'acceleration_eval_gpu.mako')
        template = Template(filename=path)
        main = template.render(helper=self)
        if self.backend == 'opencl':
            from pyopencl._cluda import CLUDA_PREAMBLE
        elif self.backend == 'cuda':
            from pycuda._cluda import CLUDA_PREAMBLE
        double_support = get_config().use_double
        cluda = Template(CLUDA_PREAMBLE).render(double_support=double_support)
        main = "\n".join([cluda, main])
        return main

    def setup_compiled_module(self, module):
        object = self.object
        self._setup_arrays_on_device()
        self.calls = self._setup_calls()
        if self.backend == 'opencl':
            acceleration_eval = GPUAccelerationEval(self)
        elif self.backend == 'cuda':
            acceleration_eval = CUDAAccelerationEval(self)

        object.set_compiled_object(acceleration_eval)

    def compile(self, code):
        if self.backend == 'opencl':
            ext = '.cl'
            backend = 'OpenCL'
        elif self.backend == 'cuda':
            ext = '.cu'
            backend = 'CUDA'
        code = convert_to_float_if_needed(code)
        path = os.path.expanduser(os.path.join(
            '~', '.pysph', 'source', get_platform_dir()
        ))
        if not os.path.exists(path):
            os.makedirs(path)
        md5 = get_md5(code)
        fname = os.path.join(path, 'm_{0}{1}'.format(md5, ext))
        # If the file already exists, we use it, this allows to write our
        # own edited version if we wish to.
        if os.path.exists(fname):
            msg = 'Reading code from {}'.format(fname)
            logger.info(msg)
            print(msg)
            with open(fname, 'r') as fp:
                code = fp.read()
        else:
            with open(fname, 'w') as fp:
                fp.write(code)
            msg = "{backend} code written to {fname}".format(
                backend=backend, fname=fname
            )
            print(msg)
            logger.info(msg)

        if self.backend == 'opencl':
            import pyopencl as cl
            self.program = cl.Program(self._ctx, code).build(
                options=['-w']
            )
        elif self.backend == 'cuda':
            from compyle.cuda import SourceModule
            self.program = SourceModule(code)
        return self.program

    ##########################################################################
    # Mako interface.
    ##########################################################################
    def get_header(self):
        object = self.object
        Converter = get_converter(self.backend)
        transpiler = Converter(known_types=self.known_types)

        headers = []
        helpers = []
        if hasattr(object.kernel, '_get_helpers_'):
            helpers.extend(object.kernel._get_helpers_())
        for equation in object.all_group.equations:
            if hasattr(equation, '_get_helpers_'):
                for helper in equation._get_helpers_():
                    if helper not in helpers:
                        helpers.append(helper)
        headers.extend(get_helper_code(helpers, transpiler, self.backend))
        headers.append(transpiler.parse_instance(object.kernel))
        cls_name = object.kernel.__class__.__name__
        self.known_types['SPH_KERNEL'] = KnownType(
            'GLOBAL_MEM %s*' % cls_name, base_type=cls_name
        )
        headers.append(object.all_group.get_equation_wrappers(
            self.known_types
        ))
        # This is to be done after the above as the equation names are assigned
        # only at this point.
        cpu_structs = self._cpu_structs
        h = CStructHelper(object.kernel)
        cpu_structs['kern'] = h.get_array()
        for eq in object.all_group.equations:
            self._equations[eq.var_name] = eq
            h.parse(eq)
            cpu_structs[eq.var_name] = h.get_array()
        return '\n'.join(headers)

    def _get_arg_base_types(self, args):
        base_types = []
        for arg in args:
            base_types.append(
                ocl_detect_pointer_base_type(arg, self.known_types.get(arg))
            )
        return base_types

    def _get_typed_args(self, args):
        code = []
        for arg in args:
            type = ocl_detect_type(arg, self.known_types.get(arg))
            code.append('{type} {arg}'.format(
                type=type, arg=arg
            ))

        return code

    def _clean_kernel_args(self, args):
        remove = ('d_idx', 's_idx')
        for a in remove:
            if a in args:
                args.remove(a)

    def _get_simple_kernel(self, g_idx, sg_idx, group, dest, all_eqs, kind,
                           source=None):
        assert kind in ('initialize', 'initialize_pair', 'post_loop', 'loop')
        sub_grp = '' if sg_idx == -1 else 's{idx}'.format(idx=sg_idx)
        if source is None:
            kernel = 'g{g_idx}{sub}_{dest}_{kind}'.format(
                g_idx=g_idx, sub=sub_grp, dest=dest, kind=kind
            )
        else:
            kernel = 'g{g_idx}{sg}_{source}_on_{dest}_{kind}'.format(
                g_idx=g_idx, sg=sub_grp, source=source, dest=dest, kind=kind
            )

        sph_k_name = self.object.kernel.__class__.__name__
        code = [
            'int d_idx = GID_0 * LDIM_0 + LID_0 + OFFSET_IDX;',
            '/* Guard for padded threads. */',
            'if (d_idx > NP_MAX) {return;};',
            'GLOBAL_MEM %s* SPH_KERNEL = kern;' % sph_k_name
        ]
        all_args, py_args, _calls = self._get_equation_method_calls(
            all_eqs, kind, indent=''
        )
        code.extend(_calls)

        s_ary, d_ary = all_eqs.get_array_names()
        if source is None:
            # We only need the dest arrays here as these are simple kernels
            # without a loop so there is no "source".
            _args = sorted(d_ary)
        else:
            d_ary.update(s_ary)
            _args = sorted(d_ary)
        py_args.extend(_args)
        all_args.extend(self._get_typed_args(_args))
        all_args.extend(
            ['GLOBAL_MEM {kernel}* kern'.format(kernel=sph_k_name),
             'double t', 'double dt', 'unsigned int NP_MAX',
             'unsigned int OFFSET_IDX']
        )

        body = '\n'.join([' ' * 4 + x for x in code])
        self.data.append(dict(
            kernel=kernel, args=py_args, dest=dest, loop=False,
            real=group.real, start_idx=group.start_idx,
            stop_idx=group.stop_idx, type='kernel'
        ))

        sig = get_kernel_definition(kernel, all_args)
        return (
            '{sig}\n{{\n{body}\n}}'.format(
                sig=sig, body=body
            )
        )

    def _get_equation_method_calls(self, eq_group, kind, indent=''):
        all_args = []
        py_args = []
        code = []
        for eq in eq_group.equations:
            method = getattr(eq, kind, None)
            if method is not None:
                cls = eq.__class__.__name__
                arg = 'GLOBAL_MEM {cls}* {name}'.format(
                    cls=cls, name=eq.var_name
                )
                all_args.append(arg)
                py_args.append(eq.var_name)
                call_args = list(getfullargspec(method).args)
                if 'self' in call_args:
                    call_args.remove('self')
                call_args.insert(0, eq.var_name)

                code.extend(
                    wrap_code(
                        '{cls}_{kind}({args});'.format(
                            cls=cls, kind=kind, args=', '.join(call_args)
                        ),
                        indent=indent
                    )
                )

        return all_args, py_args, code

    def _declare_precomp_vars(self, context):
        decl = []
        names = sorted(context.keys())
        for var in names:
            value = context[var]
            if isinstance(value, int):
                declare = 'long '
                decl.append('{declare}{var} = {value};'.format(
                    declare=declare, var=var, value=value
                ))
            elif isinstance(value, float):
                declare = 'double '
                decl.append('{declare}{var} = {value};'.format(
                    declare=declare, var=var, value=value
                ))
            elif isinstance(value, (list, tuple)):
                decl.append(
                    'double {var}[{size}];'.format(
                        var=var, size=len(value)
                    )
                )
        return decl

    def _set_kernel(self, code, kernel):
        if kernel is not None:
            name = kernel.__class__.__name__
            kern = '%s_kernel(kern, ' % name
            grad = '%s_gradient(kern, ' % name
            grad_h = '%s_gradient_h(kern, ' % name
            deltap = '%s_get_deltap(kern)' % name

            code = code.replace('DELTAP', deltap).replace('GRADIENT(', grad)
            return code.replace('KERNEL(', kern).replace('GRADH(', grad_h)
        else:
            return code

    def call_post(self, group):
        self.data.append(dict(callable=group.post, type='pre_post', args=()))

    def call_pre(self, group):
        self.data.append(dict(callable=group.pre, type='pre_post', args=()))

    def call_py_initialize(self, all_eq_group, dest):
        calls = []
        for eq in all_eq_group.equations:
            method = getattr(eq, 'py_initialize', None)
            if method is not None:
                calls.append(method)
        if len(calls) > 0:
            self.data.append(
                dict(calls=calls, type='py_initialize', dest=dest)
            )

    def call_reduce(self, all_eq_group, dest):
        self.data.append(dict(method='do_reduce', type='method',
                              args=[all_eq_group, dest]))

    def call_update_nnps(self, group):
        self.data.append(dict(method='update_nnps',
                              type='method', args=[]))

    def get_initialize_kernel(self, g_idx, sg_idx, group, dest, all_eqs):
        return self._get_simple_kernel(
            g_idx, sg_idx, group, dest, all_eqs, kind='initialize'
        )

    def get_initialize_pair_kernel(self, g_idx, sg_idx, group, dest, source,
                                   eq_group):
        return self._get_simple_kernel(
            g_idx, sg_idx, group, dest, eq_group, kind='initialize_pair',
            source=source
        )

    def get_simple_loop_kernel(self, g_idx, sg_idx, group, dest, all_eqs):
        return self._get_simple_kernel(
            g_idx, sg_idx, group, dest, all_eqs, kind='loop'
        )

    def get_post_loop_kernel(self, g_idx, sg_idx, group, dest, all_eqs):
        return self._get_simple_kernel(
            g_idx, sg_idx, group, dest, all_eqs, kind='post_loop'
        )

    def get_loop_kernel(self, g_idx, sg_idx, group, dest, source, eq_group):
        if get_config().use_local_memory:
            return self.get_lmem_loop_kernel(g_idx, sg_idx, group,
                                             dest, source, eq_group)
        kind = 'loop'
        sub_grp = '' if sg_idx == -1 else 's{idx}'.format(idx=sg_idx)
        kernel = 'g{g_idx}{sg}_{source}_on_{dest}_loop'.format(
            g_idx=g_idx, sg=sub_grp, source=source, dest=dest
        )
        sph_k_name = self.object.kernel.__class__.__name__
        context = eq_group.context
        all_args, py_args = [], []
        code = self._declare_precomp_vars(context)
        code.extend([
            'unsigned int d_idx = GID_0 * LDIM_0 + LID_0 + OFFSET_IDX;',
            '/* Guard for padded threads. */',
            'if (d_idx > NP_MAX) {return;};',
            'unsigned int s_idx, i;',
            'GLOBAL_MEM %s* SPH_KERNEL = kern;' % sph_k_name,
            'unsigned int start = start_idx[d_idx];',
            'GLOBAL_MEM unsigned int* NBRS = &(neighbors[start]);',
            'int N_NBRS = nbr_length[d_idx];',
            'unsigned int end = start + N_NBRS;'
        ])
        if eq_group.has_loop_all():
            _all_args, _py_args, _calls = self._get_equation_method_calls(
                eq_group, kind='loop_all', indent=''
            )
            code.extend(['', '// Calling loop_all of equations.'])
            code.extend(_calls)
            code.append('')
            all_args.extend(_all_args)
            py_args.extend(_py_args)

        if eq_group.has_loop():
            code.append('// Calling loop of equations.')
            code.append('for (i=start; i<end; i++) {')
            code.append('    s_idx = neighbors[i];')
            pre = []
            for p, cb in eq_group.precomputed.items():
                src = cb.code.strip().splitlines()
                pre.extend([' ' * 4 + x + ';' for x in src])
            if len(pre) > 0:
                pre.append('')
            code.extend(pre)

            _all_args, _py_args, _calls = self._get_equation_method_calls(
                eq_group, kind, indent='    '
            )
            code.extend(_calls)
            for arg, py_arg in zip(_all_args, _py_args):
                if arg not in all_args:
                    all_args.append(arg)
                    py_args.append(py_arg)
            code.append('}')

        s_ary, d_ary = eq_group.get_array_names()
        s_ary.update(d_ary)

        _args = sorted(s_ary)
        py_args.extend(_args)
        all_args.extend(self._get_typed_args(_args))

        body = '\n'.join([' ' * 4 + x for x in code])
        body = self._set_kernel(body, self.object.kernel)

        all_args.extend(
            ['GLOBAL_MEM {kernel}* kern'.format(kernel=sph_k_name),
             'GLOBAL_MEM unsigned int *nbr_length',
             'GLOBAL_MEM unsigned int *start_idx',
             'GLOBAL_MEM unsigned int *neighbors',
             'double t', 'double dt', 'unsigned int NP_MAX',
             'unsigned int OFFSET_IDX']
        )

        self.data.append(dict(
            kernel=kernel, args=py_args, dest=dest, source=source, loop=True,
            real=group.real, start_idx=group.start_idx,
            stop_idx=group.stop_idx, type='kernel'
        ))

        sig = get_kernel_definition(kernel, all_args)
        return (
            '{sig}\n{{\n{body}\n\n}}\n'.format(
                sig=sig, body=body
            )
        )

    def get_lmem_loop_kernel(self, g_idx, sg_idx, group, dest, source,
                             eq_group):
        kind = 'loop'
        sub_grp = '' if sg_idx == -1 else 's{idx}'.format(idx=sg_idx)
        kernel = 'g{g_idx}{sg}_{source}_on_{dest}_loop'.format(
            g_idx=g_idx, sg=sub_grp, source=source, dest=dest
        )
        sph_k_name = self.object.kernel.__class__.__name__
        context = eq_group.context
        all_args, py_args = [], []
        setup_code = self._declare_precomp_vars(context)
        setup_code.append('GLOBAL_MEM %s* SPH_KERNEL = kern;' % sph_k_name)

        if eq_group.has_loop_all():
            raise NotImplementedError("loop_all not suported with local "
                                      "memory")
        if group.start_idx > 0 or group.stop_idx is not None:
            raise NotImplementedError(
                "start/stop_idx not supported with local memory."
            )

        loop_code = []
        pre = []
        for p, cb in eq_group.precomputed.items():
            src = cb.code.strip().splitlines()
            pre.extend([' ' * 4 + x + ';' for x in src])
        if len(pre) > 0:
            pre.append('')
        loop_code.extend(pre)

        _all_args, _py_args, _calls = self._get_equation_method_calls(
            eq_group, kind, indent='    '
        )
        loop_code.extend(_calls)
        for arg, py_arg in zip(_all_args, _py_args):
            if arg not in all_args:
                all_args.append(arg)
                py_args.append(py_arg)

        s_ary, d_ary = eq_group.get_array_names()

        source_vars = set(s_ary)
        source_var_types = self._get_arg_base_types(source_vars)

        def modify_var_name(x):
            if x.startswith('s_'):
                return x + '_global'
            else:
                return x

        s_ary.update(d_ary)

        _args = sorted(s_ary)
        py_args.extend(_args)

        _args_modified = [modify_var_name(x) for x in _args]
        all_args.extend(self._get_typed_args(_args_modified))

        setup_body = '\n'.join([' ' * 4 + x for x in setup_code])
        setup_body = self._set_kernel(setup_body, self.object.kernel)

        loop_body = '\n'.join([' ' * 4 + x for x in loop_code])
        loop_body = self._set_kernel(loop_body, self.object.kernel)

        all_args.extend(
            ['GLOBAL_MEM {kernel}* kern'.format(kernel=sph_k_name),
             'double t', 'double dt']
        )
        all_args.extend(get_kernel_args_list())

        self.data.append(dict(
            kernel=kernel, args=py_args, dest=dest, source=source, loop=True,
            real=group.real, start_idx=group.start_idx,
            stop_idx=group.stop_idx, type='kernel'
        ))

        body = generate_body(setup=setup_body, loop=loop_body,
                             vars=source_vars, types=source_var_types,
                             wgs=get_config().wgs)

        sig = get_kernel_definition(kernel, all_args)
        return (
            '{sig}\n{{\n{body}\n\n}}\n'.format(
                sig=sig, body=body
            )
        )

    def check_condition(self, group):
        self.data.append(dict(
            type='check_condition', group=group
        ))

    def end_group(self, group):
        self.data.append(dict(
            type='end_group', group=group
        ))

    def start_iteration(self, group):
        self.data.append(dict(
            type='start_iteration', group=group
        ))

    def stop_iteration(self, group):
        self.data.append(dict(
            type='stop_iteration', group=group,
        ))



import sys
from mako.template import Template

disable_unicode = False if sys.version_info.major > 2 else True

NNPS_TEMPLATE = r"""

    /*
     * Property owners
     * octree_dst: cids, unique_cids
     * octree_src: neighbor_cid_offset, neighbor_cids
     * self evident: xsrc, ysrc, zsrc, hsrc,
                     xdst, ydst, zdst, hdst,
                     pbounds_src, pbounds_dst,
     */
    long i = get_global_id(0);
    int lid = get_local_id(0);
    int _idx = get_group_id(0);

    // Fetch dst particles
    ${data_t} _xd, _yd, _zd, _hd;

    int _cid_dst = _unique_cids[_idx];
    uint2 _pbound_here = _pbounds_dst[_cid_dst];
    char _svalid = (_pbound_here.s0 + lid < _pbound_here.s1);
    unsigned int d_idx;

    if (_svalid) {
        % if sorted:
        d_idx = _pbound_here.s0 + lid;
        % else:
        d_idx = _pids_dst[_pbound_here.s0 + lid];
        % endif

        _xd = xd[d_idx];
        _yd = yd[d_idx];
        _zd = zd[d_idx];
        _hd = hd[d_idx];
    }

    // Set loop parameters
    int _cid_src, _pid_src;
    int _offset_src = _neighbor_cid_offset[_idx];
    int _offset_lim = _neighbor_cid_offset[_idx + 1];
    uint2 _pbound_here2;

    LOCAL_MEM ${data_t} _xs[${wgs}];
    LOCAL_MEM ${data_t} _ys[${wgs}];
    LOCAL_MEM ${data_t} _zs[${wgs}];
    LOCAL_MEM ${data_t} _hs[${wgs}];

    % for var, type in zip(vars, types):
    LOCAL_MEM ${type} ${var}[${wgs}];
    % endfor

    char _nbrs[${wgs}];
    int _nbr_cnt, _m;

    ${setup}

    while (_offset_src < _offset_lim) {
        _cid_src = _neighbor_cids[_offset_src];
        _pbound_here2 = _pbounds_src[_cid_src];

        while (_pbound_here2.s0 < _pbound_here2.s1) {
            _m = min(_pbound_here2.s1,
                _pbound_here2.s0 + ${wgs}) - _pbound_here2.s0;

            // Copy src data
            if (lid < _m) {

                % if sorted:
                _pid_src = _pbound_here2.s0 + lid;
                % else:
                _pid_src = _pids_src[_pbound_here2.s0 + lid];
                % endif

                _xs[lid] = xs[_pid_src];
                _ys[lid] = ys[_pid_src];
                _zs[lid] = zs[_pid_src];
                _hs[lid] = hs[_pid_src];

                % for var in vars:
                ${var}[lid] = ${var}_global[_pid_src];
                % endfor
            }
            local_barrier();

            // Everything this point forward is done independently
            // by each thread.
            if (_svalid) {
                _nbr_cnt = 0;
                for (int _j=0; _j < _m; _j++) {
                    ${data_t} _dist2 = NORM2(_xs[_j] - _xd,
                                _ys[_j] - _yd,
                                _zs[_j] - _zd);

                    ${data_t} _r2 = MAX(_hs[_j], _hd) * _radius_scale;
                    _r2 *= _r2;

                    if (_dist2 < _r2) {
                        _nbrs[_nbr_cnt++] = _j;
                    }
                }

                int _j = 0;
                while (_j < _nbr_cnt) {
                    int s_idx = _nbrs[_j];
                    ${loop_code}
                    _j++;
                }
            }
            local_barrier();
            _pbound_here2.s0 += ${wgs};
        }

        // Process next neighboring cell
        _offset_src++;
    }

"""

NNPS_ARGS_TEMPLATE = """
    GLOBAL_MEM int *_unique_cids, GLOBAL_MEM int *_pids_src, GLOBAL_MEM int
    *_pids_dst,
    GLOBAL_MEM int *_cids,
    GLOBAL_MEM uint2 *_pbounds_src, GLOBAL_MEM uint2 *_pbounds_dst,
    %(data_t)s _radius_scale,
    GLOBAL_MEM int *_neighbor_cid_offset, GLOBAL_MEM int *_neighbor_cids,
    GLOBAL_MEM %(data_t)s *xd, GLOBAL_MEM %(data_t)s *yd,
    GLOBAL_MEM %(data_t)s *zd, GLOBAL_MEM %(data_t)s *hd,
    GLOBAL_MEM %(data_t)s *xs, GLOBAL_MEM %(data_t)s *ys,
    GLOBAL_MEM %(data_t)s *zs, GLOBAL_MEM %(data_t)s *hs
    """


def _generate_nnps_code(sorted, wgs, setup, loop, vars, types,
                        data_t='float'):
    # Note: Properties like the data type and sortedness
    # need to be fixed throughout the simulation since
    # currently this function is only called at the start of
    # the simulation.
    return Template(NNPS_TEMPLATE, disable_unicode=disable_unicode).render(
        data_t=data_t, sorted=sorted, wgs=wgs, setup=setup, loop_code=loop,
        vars=vars, types=types
    )


def generate_body(setup, loop, vars, types, wgs, c_type='float'):
    return _generate_nnps_code(True, wgs, setup, loop, vars, types,
                               c_type)


def get_kernel_args_list(c_type='float'):
    args = NNPS_ARGS_TEMPLATE % {'data_t': c_type}
    return args.split(",")
